# yapf: disable
import logging
import numpy as np
from prettytable import PrettyTable
from typing import List, Tuple, Union

from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.utils.geometry import compute_similarity_transform
from xrmocap.utils.mvpose_utils import (
    check_limb_is_correct, compute_mpjpe, vectorize_distance,
)

# yapf: enable


def evaluate(pred_keypoints3d: Keypoints,
             gt_keypoints3d: Keypoints,
             pck_thres: List = [50, 100],
             scale=1000.,
             logger: Union[None, str, logging.Logger] = None) -> dict:
    """evaluation of accuracy pred_keypoints3d (Keypoints):

    prediction of keypoints
    gt_keypoints3d (Keypoints):
        ground true of keypoints
    pck_thres (List):
        threshold value of precision
    scale:
    """
    # There must be no np.nan in the pred_keypoints3d
    mpjpe, pa_mpjpe = [], []
    pck = {i: [] for i in pck_thres}
    n_frame = gt_keypoints3d.get_frame_number()
    gt_kps3d = gt_keypoints3d.get_keypoints()[..., :3]
    gt_kps3d_mask = gt_keypoints3d.get_mask()
    pred_kps3d = pred_keypoints3d.get_keypoints()[..., :3]
    pred_kps3d_mask = pred_keypoints3d.get_mask()
    pred_kps3d_convention = pred_keypoints3d.get_convention()
    gt_kps3d_convention = gt_keypoints3d.get_convention()
    for frame_idx in range(n_frame):
        if not gt_kps3d_mask[frame_idx].any():
            continue
        gt_kps3d_idxs = np.where(
            np.sum(gt_kps3d_mask[frame_idx], axis=1) > 0)[0]
        for gt_kps3d_idx in gt_kps3d_idxs:
            f_gt_kps3d = gt_kps3d[frame_idx][gt_kps3d_idx]
            f_pred_kps3d = pred_kps3d[frame_idx][
                np.sum(pred_kps3d_mask[frame_idx], axis=1) > 0]
            if len(f_pred_kps3d) == 0:
                continue

            dist = vectorize_distance(f_gt_kps3d[np.newaxis], f_pred_kps3d)
            f_pred_kps3d = f_pred_kps3d[np.argmin(dist[0])]

            if np.all((f_pred_kps3d == 0)):
                continue

            # MPJPE
            f_pred_keypoints = Keypoints(
                kps=np.concatenate(
                    (f_pred_kps3d, np.ones_like(f_pred_kps3d[..., 0:1])),
                    axis=-1),
                convention=pred_kps3d_convention)
            f_gt_keypoints = Keypoints(
                kps=np.concatenate(
                    (f_gt_kps3d, np.ones_like(f_gt_kps3d[..., 0:1])), axis=-1),
                convention=gt_kps3d_convention)
            mpjpe.append(
                compute_mpjpe(f_pred_keypoints, f_gt_keypoints, align=True))

            # PA-MPJPE
            _, _, rotation, scaling, transl = compute_similarity_transform(
                f_gt_kps3d, f_pred_kps3d, compute_optimal_scale=True)
            pred_kps3d_pa = (scaling * f_pred_kps3d.dot(rotation)) + transl

            pred_keypoints_pa = Keypoints(
                kps=np.concatenate(
                    (pred_kps3d_pa, np.ones_like(pred_kps3d_pa[..., 0:1])),
                    axis=-1),
                convention=pred_kps3d_convention)
            pa_mpjpe_i = compute_mpjpe(
                pred_keypoints_pa, f_gt_keypoints, align=True)
            pa_mpjpe.append(pa_mpjpe_i)

            for thres in pck_thres:
                pck[thres].append(np.mean(pa_mpjpe_i <= (thres / scale)))

    mpjpe = np.asarray(mpjpe) * scale  # m to mm
    pa_mpjpe = np.asarray(pa_mpjpe) * scale  # m to mm
    mpjpe_mean, mpjpe_std = np.mean(mpjpe), np.std(mpjpe)
    pa_mpjpe_mean, pa_mpjpe_std = np.mean(pa_mpjpe), np.std(pa_mpjpe)
    # percentage
    for thres in pck_thres:
        pck[thres] = np.mean(pck[thres]) * 100.
    return dict(
        mpjpe_mean=mpjpe_mean,
        mpjpe_std=mpjpe_std,
        pa_mpjpe_mean=pa_mpjpe_mean,
        pa_mpjpe_std=pa_mpjpe_std,
        pck=pck)


def calc_limbs_accuracy(
    pred_keypoints3d,
    gt_keypoints3d,
    limbs,
    logger: Union[None, str, logging.Logger] = None
) -> Tuple[np.ndarray, PrettyTable]:
    """calculate the limbs accuracy pred_keypoints3d (Keypoints):

    prediction of keypoints
    gt_keypoints3d (Keypoints):
        ground true of keypoints
    limbs:
        limb to be evaluated
    """
    n_frame = gt_keypoints3d.get_frame_number()
    n_gt_person = gt_keypoints3d.get_person_number()
    gt_kps3d = gt_keypoints3d.get_keypoints()[..., :3]
    gt_kps3d_mask = gt_keypoints3d.get_mask()
    pred_kps3d = pred_keypoints3d.get_keypoints()[..., :3]
    pred_kps3d_mask = pred_keypoints3d.get_mask()
    check_result = np.zeros((n_frame, n_gt_person, len(limbs) + 1),
                            dtype=np.int32)
    accuracy_cnt = 0
    error_cnt = 0

    for idx in range(n_frame):
        if not gt_kps3d_mask[idx].any():
            continue
        gt_kps3d_idxs = np.where(np.sum(gt_kps3d_mask[idx], axis=1) > 0)[0]
        for gt_kps3d_idx in gt_kps3d_idxs:
            f_gt_kps3d = gt_kps3d[idx][gt_kps3d_idx]
            f_pred_kps3d = pred_kps3d[idx][
                np.sum(pred_kps3d_mask[idx], axis=1) > 0]
            if len(f_pred_kps3d) == 0:
                continue

            dist = vectorize_distance(f_gt_kps3d[np.newaxis], f_pred_kps3d)
            f_pred_kps3d = f_pred_kps3d[np.argmin(dist[0])]

            for i, limb in enumerate(limbs):
                start_point, end_point = limb
                if check_limb_is_correct(f_pred_kps3d[start_point],
                                         f_pred_kps3d[end_point],
                                         f_gt_kps3d[start_point],
                                         f_gt_kps3d[end_point]):
                    check_result[idx, gt_kps3d_idx, i] = 1
                    accuracy_cnt += 1
                else:
                    check_result[idx, gt_kps3d_idx, i] = -1
                    error_cnt += 1
            gt_hip = (f_gt_kps3d[2] + f_gt_kps3d[3]) / 2
            pred_hip = (f_pred_kps3d[2] + f_pred_kps3d[3]) / 2
            if check_limb_is_correct(pred_hip, f_pred_kps3d[12], gt_hip,
                                     f_gt_kps3d[12]):
                check_result[idx, gt_kps3d_idx, -1] = 1
                accuracy_cnt += 1
            else:
                check_result[idx, gt_kps3d_idx, -1] = -1
                error_cnt += 1
    bone_group = dict([('Torso', np.array([len(limbs) - 1])),
                       ('Upper arms', np.array([5, 6])),
                       ('Lower arms', np.array([4, 7])),
                       ('Upper legs', np.array([1, 2])),
                       ('Lower legs', np.array([0, 3]))])
    if len(limbs) > 9:
        # head is absent in some dataset
        bone_group['Head'] = np.array([8])

    person_wise_avg = np.sum(
        check_result > 0, axis=(0, 2)) / np.sum(
            np.abs(check_result), axis=(0, 2))

    bone_wise_result = dict()
    bone_person_wise_result = dict()
    for k, v in bone_group.items():
        bone_wise_result[k] = np.sum(check_result[:, :, v] > 0) / np.sum(
            np.abs(check_result[:, :, v]))
        bone_person_wise_result[k] = np.sum(
            check_result[:, :, v] > 0, axis=(0, 2)) / np.sum(
                np.abs(check_result[:, :, v]), axis=(0, 2))

    tb = PrettyTable()
    tb.field_names = ['Bone Group'] + [
        f'Actor {i}' for i in range(bone_person_wise_result['Torso'].shape[0])
    ] + ['Average']
    for k, v in bone_person_wise_result.items():
        this_row = [k] + [np.char.mod('%.4f', i) for i in v
                          ] + [np.char.mod('%.4f',
                                           np.sum(v) / len(v))]
        tb.add_row(this_row)
    this_row = ['Total'] + [
        np.char.mod('%.4f', i) for i in person_wise_avg
    ] + [np.char.mod('%.4f',
                     np.sum(person_wise_avg) / len(person_wise_avg))]
    tb.add_row(this_row)
    return check_result, tb



# yapf: disable
import logging
import numpy as np
from typing import List, Tuple, Union, overload
from xrprimer.data_structure.camera import FisheyeCameraParameter
from xrprimer.utils.ffmpeg_utils import video_to_array

from xrmocap.data_structure.body_model import SMPLData
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.human_perception.builder import (
    MMdetDetector, MMposeTopDownEstimator, build_detector,
)
from xrmocap.io.image import load_multiview_images
from xrmocap.model.registrant.builder import SMPLify, build_registrant
from xrmocap.model.registrant.handler.builder import build_handler
from xrmocap.ops.triangulation.builder import (
    BaseTriangulator, build_triangulator,
)
from xrmocap.ops.triangulation.point_selection.builder import (
    BaseSelector, CameraErrorSelector, build_point_selector,
)
from xrmocap.transform.convention.keypoints_convention import convert_keypoints
from xrmocap.transform.keypoints3d.optim.builder import (
    BaseOptimizer, build_keypoints3d_optimizer,
)
from .base_estimator import BaseEstimator

# yapf: enable


class MultiViewSinglePersonSMPLEstimator(BaseEstimator):
    """Api for estimating smpl in a multi-view single-person scene."""

    def __init__(self,
                 work_dir: str,
                 bbox_detector: Union[dict, MMdetDetector],
                 kps2d_estimator: Union[dict, MMposeTopDownEstimator],
                 triangulator: Union[dict, BaseTriangulator],
                 smplify: Union[dict, SMPLify],
                 cam_pre_selector: Union[dict, BaseSelector, None] = None,
                 cam_selector: Union[dict, CameraErrorSelector, None] = None,
                 final_selectors: List[Union[dict, BaseSelector, None]] = None,
                 kps3d_optimizers: Union[List[Union[BaseOptimizer, dict]],
                                         None] = None,
                 load_batch_size: int = 500,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Initialization of the class.

        Args:
            work_dir (str):
                Path to the folder for running the api. No file in work_dir
                will be modified
                or added by MultiViewSinglePersonSMPLEstimator.
            bbox_detector (Union[dict, MMdetDetector]):
                A human bbox_detector or its config.
            kps2d_estimator (Union[dict, MMposeTopDownEstimator]):
                A top-down kps2d estimator or its config.
            triangulator (Union[dict, BaseTriangulator]):
                A triangulator or its config.
            smplify (Union[dict, SMPLify]):
                A SMPLify instance or its config.
            cam_pre_selector (Union[dict, BaseSelector, None], optional):
                A selector before selecting cameras. If it's given,
                points for camera selection will be filtered.
                Defaults to None.
            cam_selector (Union[dict, CameraErrorSelector, None], optional):
                A camera selector or its config. If it's given, cameras
                will be selected before triangulation.
                Defaults to None.
            final_selectors (List[Union[dict, BaseSelector, None]], optional):
                A list of selectors or their configs. If given, kps2d will be
                selected by the cascaded final selectors before triangulation.
                Defaults to None.
            kps3d_optimizers (Union[
                    List[Union[BaseOptimizer, dict]], None], optional):
                A list of keypoints3d optimizers or their configs. If given,
                keypoints3d will be
                optimized by the cascaded final optimizers after triangulation.
                Defaults to None.
            load_batch_size (int, optional):
                How many frames are loaded at the same time. Defaults to 500.
            verbose (bool, optional):
                Whether to print(logger.info) information during estimating.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(work_dir, verbose, logger)
        self.load_batch_size = load_batch_size

        if isinstance(bbox_detector, dict):
            bbox_detector['logger'] = logger
            self.bbox_detector = build_detector(bbox_detector)
        else:
            self.bbox_detector = bbox_detector

        if isinstance(kps2d_estimator, dict):
            kps2d_estimator['logger'] = logger
            self.kps2d_estimator = build_detector(kps2d_estimator)
        else:
            self.kps2d_estimator = kps2d_estimator

        if isinstance(triangulator, dict):
            triangulator['logger'] = logger
            self.triangulator = build_triangulator(triangulator)
        else:
            self.triangulator = triangulator

        if isinstance(smplify, dict):
            smplify['logger'] = logger
            self.smplify = build_registrant(smplify)
        else:
            self.smplify = smplify

        if isinstance(cam_pre_selector, dict):
            cam_pre_selector['logger'] = logger
            self.cam_pre_selector = build_point_selector(cam_pre_selector)
        else:
            self.cam_pre_selector = cam_pre_selector

        if isinstance(cam_selector, dict):
            cam_selector['logger'] = logger
            cam_selector['triangulator']['camera_parameters'] = \
                self.triangulator.camera_parameters
            self.cam_selector = build_point_selector(cam_selector)
        else:
            self.cam_selector = cam_selector

        if final_selectors is None:
            self.final_selectors = None
        else:
            self.final_selectors = []
            for selector in final_selectors:
                if isinstance(selector, dict):
                    selector['logger'] = logger
                    selector = build_point_selector(selector)
                self.final_selectors.append(selector)

        if kps3d_optimizers is None:
            self.kps3d_optimizers = None
        else:
            self.kps3d_optimizers = []
            for kps3d_optim in kps3d_optimizers:
                if isinstance(kps3d_optim, dict):
                    kps3d_optim['logger'] = logger
                    kps3d_optim = build_keypoints3d_optimizer(kps3d_optim)
                self.kps3d_optimizers.append(kps3d_optim)

    @overload
    def run(
        self, img_arr: np.ndarray, cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    @overload
    def run(
        self, img_paths: List[List[str]],
        cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    @overload
    def run(
        self, video_path: List[str], cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    def run(
        self,
        cam_param: List[FisheyeCameraParameter],
        img_arr: Union[None, np.ndarray] = None,
        img_paths: Union[None, List[List[str]]] = None,
        video_paths: Union[None, List[str]] = None,
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        """Run mutli-view single-person smpl estimator once. run() needs one
        images input among [img_arr, img_paths, video_paths].

        Args:
            cam_param (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameter instances.
            img_arr (Union[None, np.ndarray], optional):
                A multi-view image array, in shape
                [n_view, n_frame, h, w, c]. Defaults to None.
            img_paths (Union[None, List[List[str]]], optional):
                A nested list of image paths, in shape
                [n_view, n_frame]. Defaults to None.
            video_paths (Union[None, List[str]], optional):
                A list of video paths, each is a view.
                Defaults to None.

        Returns:
            Tuple[List[Keypoints], Keypoints, SMPLData]:
                A list of kps2d, an instance of Keypoints 3d,
                an instance of SMPLData.
        """
        input_list = [img_arr, img_paths, video_paths]
        input_count = 0
        for input_instance in input_list:
            if input_instance is not None:
                input_count += 1
        if input_count > 1:
            self.logger.error('Redundant input!\n' +
                              'Please offer only one among' +
                              ' img_arr, img_paths and video_paths.')
            raise ValueError
        if img_arr is not None:
            mview_img_arr = img_arr
        elif img_paths is not None:
            mview_img_arr = load_multiview_images(img_paths)
        elif video_paths is not None:
            mview_img_list = []
            for video_path in video_paths:
                sv_img_arr = video_to_array(input_path=video_path)
                mview_img_list.append(sv_img_arr)
            mview_img_arr = np.asarray(mview_img_list)
        else:
            self.logger.error('No image input has been found!\n' +
                              'img_arr, img_paths and video_paths are None.')
            raise ValueError
        keypoints2d_list = self.estimate_keypoints2d(img_arr=mview_img_arr)
        keypoints3d = self.estimate_keypoints3d(
            cam_param=cam_param, keypoints2d_list=keypoints2d_list)
        smpl_data = self.estimate_smpl(keypoints3d=keypoints3d)
        return keypoints2d_list, keypoints3d, smpl_data

    def estimate_keypoints2d(
        self,
        img_arr: Union[None, np.ndarray] = None,
        img_paths: Union[None, List[List[str]]] = None,
    ) -> List[Keypoints]:
        """Estimate keypoints2d in a top-down way.

        Args:
            img_arr (Union[None, np.ndarray], optional):
                A multi-view image array, in shape
                [n_view, n_frame, h, w, c]. Defaults to None.
            img_paths (Union[None, List[List[str]]], optional):
                A nested list of image paths, in shape
                [n_view, n_frame]. Defaults to None.

        Returns:
            List[Keypoints]:
                A list of keypoints2d instances.
        """
        self.logger.info('Estimating keypoints2d.')
        input_list = [img_arr, img_paths]
        input_count = 0
        for input_instance in input_list:
            if input_instance is not None:
                input_count += 1
        if input_count > 1:
            self.logger.error('Redundant input!\n' +
                              'Please offer only one between' +
                              ' img_arr, img_paths.')
            raise ValueError
        ret_list = []
        for view_index in range(img_arr.shape[0]):
            if img_arr is not None:
                view_img_arr = img_arr[view_index]
                bbox_list = self.bbox_detector.infer_array(
                    image_array=view_img_arr,
                    disable_tqdm=(not self.verbose),
                    multi_person=False)
                kps2d_list, _, _ = self.kps2d_estimator.infer_array(
                    image_array=view_img_arr,
                    bbox_list=bbox_list,
                    disable_tqdm=(not self.verbose),
                )
            else:
                bbox_list = self.bbox_detector.infer_frames(
                    frame_path_list=img_paths,
                    disable_tqdm=(not self.verbose),
                    multi_person=True,
                    load_batch_size=self.load_batch_size)
                kps2d_list, _, _ = self.kps2d_estimator.infer_frames(
                    frame_path_list=img_paths,
                    bbox_list=bbox_list,
                    disable_tqdm=(not self.verbose),
                    load_batch_size=self.load_batch_size)
            if len(kps2d_list) == 1 and \
                    len(kps2d_list[0]) == 1 and \
                    kps2d_list[0][0] is None:
                kps2d_list = [[]]
            keypoints2d = self.kps2d_estimator.get_keypoints_from_result(
                kps2d_list)
            ret_list.append(keypoints2d)
        return ret_list

    def estimate_keypoints3d(self, cam_param: List[FisheyeCameraParameter],
                             keypoints2d_list: List[Keypoints]) -> Keypoints:
        """Estimate keypoints3d by triangulation and optimizers if exists.

        Args:
            cam_param (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameter instances.
            keypoints2d_list (List[Keypoints]):
                A list of Keypoints2d, in same mask and convention,
                and the time axis are aligned.

        Returns:
            Keypoints: A keypoints3d Keypoints instance.
        """
        self.logger.info('Estimating keypoints3d.')
        # prepare input np.ndarray
        kps_arr_list = []
        mask_list = []
        default_keypoints2d = None
        for keypoints2d in keypoints2d_list:
            if keypoints2d is not None:
                default_keypoints2d = keypoints2d.clone()
                default_keypoints2d.set_keypoints(
                    np.zeros_like(default_keypoints2d.get_keypoints()))
                default_keypoints2d.set_mask(
                    np.zeros_like(default_keypoints2d.get_mask()))
                break
        if default_keypoints2d is None:
            self.logger.error('No one has been detected in any view.')
            raise AttributeError
        for keypoints2d in keypoints2d_list:
            if keypoints2d is None:
                keypoints2d = default_keypoints2d
            if keypoints2d.dtype != 'numpy':
                keypoints2d = keypoints2d.to_numpy()
            kps_arr_list.append(keypoints2d.get_keypoints()[:, 0, ...])
            mask_list.append(keypoints2d.get_mask()[:, 0, ...])
        mview_kps2d_arr = np.asarray(kps_arr_list)
        mview_mask = np.asarray(mask_list)
        mview_mask = np.expand_dims(mview_mask, -1)
        # select camera
        cam_indexes = self.select_camera(cam_param, mview_kps2d_arr,
                                         mview_mask)
        self.triangulator.set_cameras(cam_param)
        selected_triangulator = self.triangulator[cam_indexes]
        mview_kps2d_arr = mview_kps2d_arr[np.asarray(cam_indexes), ...]
        triangulate_mask = mview_mask[np.asarray(cam_indexes), ...]
        # cascade point selectors
        self.logger.info('Selecting points.')
        if self.final_selectors is not None:
            for selector in self.final_selectors:
                triangulate_mask = selector.get_selection_mask(
                    points=mview_kps2d_arr, init_points_mask=triangulate_mask)
        kps3d_arr = selected_triangulator.triangulate(
            points=mview_kps2d_arr, points_mask=triangulate_mask)
        kps3d_arr = np.concatenate(
            (kps3d_arr, np.ones_like(kps3d_arr[..., 0:1])), axis=-1)
        kps3d_arr = np.expand_dims(kps3d_arr, axis=1)
        kps3d_mask = np.sum(mview_mask, axis=(0, 1), keepdims=False)
        kps3d_mask = np.sign(np.abs(kps3d_mask))
        if kps3d_mask.shape[-1] == 1:
            kps3d_mask = kps3d_mask[..., 0]
        keypoints3d = Keypoints(
            dtype='numpy',
            kps=kps3d_arr,
            mask=kps3d_mask,
            convention=default_keypoints2d.get_convention())
        optim_kwargs = dict(
            mview_kps2d=np.expand_dims(mview_kps2d_arr, axis=2),
            mview_kps2d_mask=np.expand_dims(triangulate_mask, axis=2))
        if self.kps3d_optimizers is not None:
            for optimizer in self.kps3d_optimizers:
                if hasattr(optimizer, 'triangulator'):
                    optimizer.triangulator = selected_triangulator
                keypoints3d = optimizer.optimize_keypoints3d(
                    keypoints3d, **optim_kwargs)
        return keypoints3d

    def estimate_smpl(self, keypoints3d: Keypoints) -> SMPLData:
        """Estimate smpl parameters according to keypoints3d.

        Args:
            keypoints3d (Keypoints):
                A keypoints3d Keypoints instance, with only one person
                inside. This method will take the person at
                keypoints3d.get_keypoints()[:, 0, ...] to run smplify.

        Returns:
            SMPLData:
                Smpl data of the person.
        """
        self.logger.info('Estimating SMPL.')
        working_convention = 'smpl'
        keypoints3d = convert_keypoints(
            keypoints=keypoints3d, dst=working_convention)
        keypoints3d = keypoints3d.to_tensor(device=self.smplify.device)
        kps3d_tensor = keypoints3d.get_keypoints()[:, 0, :, :3].float()
        kps3d_conf = keypoints3d.get_mask()[:, 0, ...]
        # build and run
        kp3d_mse_input = build_handler(
            dict(
                type='Keypoint3dMSEInput',
                keypoints3d=kps3d_tensor,
                keypoints3d_conf=kps3d_conf,
                keypoints3d_convention=working_convention,
                handler_key='keypoints3d_mse'))
        kp3d_llen_input = build_handler(
            dict(
                type='Keypoint3dLimbLenInput',
                keypoints3d=kps3d_tensor,
                keypoints3d_conf=kps3d_conf,
                keypoints3d_convention=working_convention,
                handler_key='keypoints3d_limb_len'))
        registrant_output = self.smplify(
            input_list=[kp3d_mse_input, kp3d_llen_input])
        smpl_data = SMPLData()
        smpl_data.from_param_dict(registrant_output)
        return smpl_data

    def select_camera(self, cam_param: List[FisheyeCameraParameter],
                      points: np.ndarray,
                      points_mask: np.ndarray) -> List[int]:
        """Use cam_pre_selector to filter bad points, use reprojection error of
        the good points to select good cameras.

        Args:
            cam_param (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameter instances.
            points (np.ndarray):
                Multiview points2d, in shape [n_view, n_frame, n_kps, 3].
                Point scores at the last dim.
            points_mask (np.ndarray):
                Multiview points2d mask,
                in shape [n_view, n_frame, n_kps, 1].

        Returns:
            List[int]: A list of camera indexes.
        """
        if self.cam_selector is not None:
            self.logger.info('Selecting cameras.')
            if self.cam_pre_selector is not None:
                self.logger.info('Using pre-selector for camera selection.')
                pre_mask = self.cam_pre_selector.get_selection_mask(
                    points=points, init_points_mask=points_mask)
            else:
                pre_mask = points_mask.copy()
            self.triangulator.set_cameras(cam_param)
            self.cam_selector.triangulator = self.triangulator
            selected_camera_indexes = self.cam_selector.get_camera_indexes(
                points=points, init_points_mask=pre_mask)
            self.logger.info(f'Selected cameras: {selected_camera_indexes}')
        else:
            self.logger.warning(
                'The estimator api instance has no cam_selector,' +
                ' all the cameras will be returned.')
            selected_camera_indexes = [idx for idx in range(len(cam_param))]
        return selected_camera_indexes



# yapf: disable
import numpy as np
import string
from typing import List

from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.transform.convention.keypoints_convention import (
    convert_keypoints, get_keypoint_idx,
)
from xrmocap.transform.limbs import get_limbs_from_keypoints
from xrmocap.utils.mvpose_utils import (
    add_campus_jaw_headtop, add_campus_jaw_headtop_mask,
)

# yapf: enable


def align_keypoints3d(pred_keypoints3d: Keypoints, gt_keypoints3d: Keypoints,
                      eval_kps3d_convention: string,
                      selected_limbs_name: List[List[str]],
                      additional_limbs_names: List[List[str]]):
    """align keypoints convention.

    Args:
        pred_keypoints3d (Keypoints): prediction of keypoints
        gt_keypoints3d (Keypoints): ground true of keypoints
        eval_kps3d_convention (string): keypoints convention to align
        selected_limbs_name (List): selected limbs to be evaluated
        additional_limbs_names (List): additional limbs to be evaluated
    """
    ret_limbs = []
    gt_nose = None
    pred_nose = None
    pred_kps3d_convention = pred_keypoints3d.get_convention()
    gt_kps3d_convention = gt_keypoints3d.get_convention()
    if gt_kps3d_convention == 'panoptic':
        gt_nose_index = get_keypoint_idx(
            name='nose_openpose', convention=gt_kps3d_convention)
        gt_nose = gt_keypoints3d.get_keypoints()[:, :, gt_nose_index, :3]

    if pred_kps3d_convention == 'coco':
        pred_nose_index = get_keypoint_idx(
            name='nose', convention=pred_kps3d_convention)
        pred_nose = pred_keypoints3d.get_keypoints()[:, :, pred_nose_index, :3]

    if pred_kps3d_convention == 'fourdag_19' or\
            pred_kps3d_convention == 'openpose_25':
        pred_leftear_index = get_keypoint_idx(
            name='left_ear_openpose', convention=pred_kps3d_convention)
        pre_rightear_index = get_keypoint_idx(
            name='right_ear_openpose', convention=pred_kps3d_convention)
        head_center = (
            pred_keypoints3d.get_keypoints()[:, :, pred_leftear_index, :3] +
            pred_keypoints3d.get_keypoints()[:, :, pre_rightear_index, :3]) / 2
        pred_nose = head_center

    if pred_kps3d_convention != eval_kps3d_convention:
        pred_keypoints3d = convert_keypoints(
            keypoints=pred_keypoints3d,
            dst=eval_kps3d_convention,
            approximate=True)
    if gt_kps3d_convention != eval_kps3d_convention:
        gt_keypoints3d = convert_keypoints(
            keypoints=gt_keypoints3d,
            dst=eval_kps3d_convention,
            approximate=True)
    limbs = get_limbs_from_keypoints(
        keypoints=pred_keypoints3d, fill_limb_names=True)
    limb_name_list = []
    conn_list = []
    for limb_name, conn in limbs.get_connections_by_names().items():
        limb_name_list.append(limb_name)
        conn_list.append(conn)

    for idx, limb_name in enumerate(limb_name_list):
        if limb_name in selected_limbs_name:
            ret_limbs.append(conn_list[idx])

    for conn_names in additional_limbs_names:
        kps_idx_0 = get_keypoint_idx(
            name=conn_names[0], convention=eval_kps3d_convention)
        kps_idx_1 = get_keypoint_idx(
            name=conn_names[1], convention=eval_kps3d_convention)
        ret_limbs.append(np.array([kps_idx_0, kps_idx_1], dtype=np.int32))
    pred_kps3d_mask = pred_keypoints3d.get_mask()
    pred_kps3d = pred_keypoints3d.get_keypoints()[..., :3]
    if pred_nose is not None:
        pred_kps3d = add_campus_jaw_headtop(pred_nose, pred_kps3d)
        pred_kps3d_mask = add_campus_jaw_headtop_mask(pred_kps3d_mask)

    gt_kps3d_mask = gt_keypoints3d.get_mask()
    gt_kps3d = gt_keypoints3d.get_keypoints()[..., :3]
    if gt_nose is not None:
        gt_kps3d = add_campus_jaw_headtop(gt_nose, gt_kps3d)
        gt_kps3d_mask = add_campus_jaw_headtop_mask(gt_kps3d_mask)

    pred_kps3d = np.concatenate((pred_kps3d, pred_kps3d_mask[..., np.newaxis]),
                                axis=-1)
    pred_keypoints3d = Keypoints(
        kps=pred_kps3d, mask=pred_kps3d_mask, convention=eval_kps3d_convention)
    gt_kps3d = np.concatenate((gt_kps3d, gt_kps3d_mask[..., np.newaxis]),
                              axis=-1)
    gt_keypoints3d = Keypoints(
        kps=gt_kps3d, mask=gt_kps3d_mask, convention=eval_kps3d_convention)

    return pred_keypoints3d, gt_keypoints3d, ret_limbs



# yapf: disable
from .visualize_keypoints2d import visualize_keypoints2d
from .visualize_keypoints3d import (
    visualize_keypoints3d, visualize_project_keypoints3d,
)

# yapf: enable
__all__ = [
    'visualize_keypoints2d', 'visualize_keypoints3d',
    'visualize_project_keypoints3d'
]



import numpy as np
import torch


def unfold_camera_param(camera: dict):
    """This function is to extract camera extrinsic, intrinsic and distorsion
    parameters from dictionary.

    Args:
        camera (dict): Dictionary to store the camera parameters.

    Returns:
        R (Union[np.ndarray, torch.Tensor]):
            Extrinsic parameters, rotation matrix.
        T (Union[np.ndarray, torch.Tensor]):
            Extrinsic parameters, translation matrix.
        f (Union[np.ndarray, torch.Tensor]):
            Focal length in x, y direction.
        c (Union[np.ndarray, torch.Tensor]):
            Camera center.
        k (Union[list, torch.Tensor]):
            Radial distortion coefficients.
        p (Union[list, torch.Tensor]):
            Tangential distortion coefficients.
    """

    R = camera['R']
    T = camera['T']
    K = camera['K']
    dist_coeff = camera['dist_coeff']

    if not hasattr(R, 'device'):
        fx = K[0, 0]
        fy = K[1, 1]
        cx = K[0, 2]
        cy = K[1, 2]
        k = dist_coeff[[0, 1, 4]]
        p = dist_coeff[[2, 3]]
        f = np.array([[fx], [fy]]).reshape(-1, 1)
        c = np.array([[cx], [cy]]).reshape(-1, 1)
        return R, T, f, c, k, p

    else:
        device = R.device
        R = torch.as_tensor(R, dtype=torch.float, device=device)
        T = torch.as_tensor(T, dtype=torch.float, device=device)
        K = torch.as_tensor(K, dtype=torch.float, device=device)
        dist_coeff = torch.as_tensor(
            dist_coeff, dtype=torch.float, device=device)

        if R.ndim == 3:  # [bs, (cam_param)]
            R = R.reshape(3, 3)
            T = T.reshape(3, 1)
            K = K.reshape(3, 3)
            dist_coeff = dist_coeff.reshape(8, 1)
            k = dist_coeff[[0, 1, 4]]
            p = dist_coeff[[2, 3]]
            f = torch.tensor([K[0, 0], K[1, 1]],
                             dtype=torch.float,
                             device=device).reshape(2, 1)
            c = torch.as_tensor([[K[0, 2]], [K[1, 2]]],
                                dtype=torch.float,
                                device=device).reshape(2, 1)
            return R, T, f, c, k, p

        elif R.ndim == 4:  # [bs, n_views, (cam_param)]
            batch_size, n_views = R.shape[:2]
            k = dist_coeff[:, :, [0, 1, 4], None]
            p = dist_coeff[:, :, [2, 3], None]

            f = torch.tensor(
                torch.stack([K[:, :, 0, 0], K[:, :, 1, 1]], dim=2),
                dtype=torch.float,
                device=device).view(batch_size, n_views, 2, 1)
            c = torch.as_tensor(
                torch.stack([K[:, :, 0, 2], K[:, :, 1, 2]], dim=2),
                dtype=torch.float,
                device=device).view(batch_size, n_views, 2, 1)
            return R, T, f, c, k, p

        else:
            raise ValueError(f'Invalid camera parameter shape: {R.shape}')


def project_point_radial(x, R, T, f, c, k, p):
    """This function is to project a point in 3D space to 2D pixel space with
    given camera parameters.

    Args:
        x: Nx3 points in world coordinates
        R: 3x3 Camera rotation matrix
        T: 3x1 Camera translation parameters
        f: (scalar) Camera focal length
        c: 2x1 Camera center
        k: 3x1 Camera radial distortion coefficients
        p: 2x1 Camera tangential distortion coefficients

    Returns:
        ypixel.T: Nx2 points in pixel space
    """
    if not hasattr(x, 'device'):
        n = x.shape[0]
        x_cam = R.dot(x.T - T)
        y = x_cam[:2] / (x_cam[2] + 1e-5)

        r2 = np.sum(y**2, axis=0)
        radial = 1 + np.einsum('ij,ij->j', np.tile(k, (1, n)),
                               np.array([r2, r2**2, r2**3]))
        tan = p[0] * y[1] + p[1] * y[0]
        y = y * np.tile(radial + 2 * tan, (2, 1)) + np.outer(
            np.array([p[1], p[0]]).reshape(-1), r2)
        ypixel = np.multiply(f, y) + c
        return ypixel.T

    else:
        if x.ndim == 2:
            n = x.shape[0]
            x_cam = torch.mm(R, torch.t(x) - T)
            y = x_cam[:2] / (x_cam[2] + 1e-5)

            kexp = k.repeat((1, n))
            r2 = torch.sum(y**2, 0, keepdim=True)
            r2exp = torch.cat([r2, r2**2, r2**3], 0)
            radial = 1 + torch.einsum('ij,ij->j', kexp, r2exp)

            tan = p[0] * y[1] + p[1] * y[0]
            corr = (radial + 2 * tan).repeat((2, 1))

            y = y * corr + torch.ger(
                torch.cat([p[1], p[0]]).view(-1), r2.view(-1))
            ypixel = (f * y) + c
            return torch.t(ypixel)

        elif x.ndim == 4:
            bs, n_view, n_bins, _ = x.shape
            x_cam = torch.matmul(R, x.transpose(2, 3) - T)

            y = x_cam[:, :, :2] / (x_cam[:, :, 2:] + 1e-5)

            kexp = k.repeat(1, 1, 1, n_bins)
            r2 = torch.sum(y**2, 2, keepdim=True)
            r2exp = torch.cat([r2, r2**2, r2**3], 2)
            radial = 1 + torch.einsum('bvij,bvij->bvj', kexp, r2exp)

            tan = p[:, :, 0] * y[:, :, 1] + p[:, :, 1] * y[:, :, 0]
            corr = (radial + 2 * tan).unsqueeze(2).expand(-1, -1, 2, -1)

            y = (
                y * corr +
                torch.matmul(torch.stack([p[:, :, 1], p[:, :, 0]], dim=2), r2))
            ypixel = (f * y) + c
            return ypixel.transpose(2, 3)


def project_pose(x, camera):
    R, T, f, c, k, p = unfold_camera_param(camera)
    return project_point_radial(x, R, T, f, c, k, p)



import logging
import numpy as np
import torch
from typing import Union

from .base_image_transform import BaseImageTransform


class BGR2RGB(BaseImageTransform):

    def __init__(self,
                 color_dim: int = -1,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Convert image array of any shape between BGR and RGB.

        Args:
            color_dim (int, optional):
                Which dim is the color channel. Defaults to -1.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        BaseImageTransform.__init__(self, logger=logger)
        self.color_dim = color_dim

    def forward(
        self, input: Union[np.ndarray,
                           torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
        """Forward function of BGR2RGB.

        Args:
            input (Union[np.ndarray, torch.Tensor]):
                An array of images. The shape could be:
                [h, w, n_ch], [n_frame, h, w, n_ch],
                [n_view, n_frame, h, w, n_ch], etc.

        Returns:
            Union[np.ndarray, torch.Tensor]
        """
        return bgr2rgb(input, self.color_dim)


def bgr2rgb(input_array: Union[np.ndarray, torch.Tensor],
            color_dim: int = -1) -> Union[np.ndarray, torch.Tensor]:
    """Convert image array of any shape between BGR and RGB.

    Args:
        input_array (Union[np.ndarray, torch.Tensor]):
            An array of images. The shape could be:
            [h, w, n_ch], [n_frame, h, w, n_ch],
            [n_view, n_frame, h, w, n_ch], etc.
        color_dim (int, optional):
            Which dim is the color channel. Defaults to -1.

    Returns:
        Union[np.ndarray, torch.Tensor]:
            Same type as the input.
    """
    r_slice_list = [
        slice(None),
    ] * len(input_array.shape)
    b_slice_list = [
        slice(None),
    ] * len(input_array.shape)
    r_slice_list[color_dim] = slice(0, 1, 1)
    b_slice_list[color_dim] = slice(2, 3, 1)
    if isinstance(input_array, torch.Tensor):
        b_backup = input_array[tuple(b_slice_list)].clone()
    else:
        b_backup = input_array[tuple(b_slice_list)].copy()
    input_array[tuple(b_slice_list)] = input_array[tuple(r_slice_list)]
    input_array[tuple(r_slice_list)] = b_backup
    return input_array



# yapf: disable
from xrmocap.human_perception.bbox_detection.mmdet_detector import (
    MMdetDetector,
)
from xrmocap.human_perception.builder import DETECTORS
from xrmocap.human_perception.keypoints_estimation.mmpose_top_down_estimator import \
    MMposeTopDownEstimator  # noqa:E501

# yapf: enable

__all__ = ['DETECTORS', 'MMdetDetector', 'MMposeTopDownEstimator']



# yapf: disable
import logging
import numpy as np
import torch
from smplx import SMPLX as _SMPLX
from smplx.lbs import vertices2joints
from typing import Union
from xrprimer.utils.log_utils import get_logger

from xrmocap.transform.convention.keypoints_convention import (
    get_keypoints_factory,
)

# yapf: enable


class SMPLX(_SMPLX):
    NUM_VERTS = 10475
    NUM_FACES = 20908

    def __init__(self,
                 *args,
                 use_face_contour: bool = True,
                 use_pca: bool = False,
                 flat_hand_mean: bool = True,
                 keypoint_convention: str = 'smplx',
                 joints_regressor: str = None,
                 extra_joints_regressor: str = None,
                 logger: Union[None, str, logging.Logger] = None,
                 **kwargs) -> None:
        """Extension of the official SMPLX implementation.

        Args:
            *args:
                extra arguments for SMPLX initialization.
            use_face_contour (bool, optional):
                Whether to compute the keypoints that form the facial contour.
                If selected, there will be 144 keypoints.
                Defaults to True.
            use_pca (bool, optional):
                Whether to use pca for hand poses.
                Defaults to False.
            flat_hand_mean (bool, optional):
                If False, then the pose of the hand is initialized to False.
                Defaults to True.
            keypoint_convention (str, optional):
                Source convention of keypoints. This convention
                is used for keypoints obtained from joint regressors.
                Defaults to 'smplx'.
            joints_regressor (str, optional):
                Path to joint regressor. Should be an npy file.
                If provided, replaces the official J_regressor of SMPLX.
                Defaults to None.
            extra_joints_regressor (str, optional):
                Path to extra joint regressor. Should be
                an npy file. If provided, extra joints are regressed and
                concatenated after the joints regressed with the official
                J_regressor or joints_regressor.
                Defaults to None.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
            **kwargs:
                extra keyword arguments for SMPLX initialization.
        """
        super(SMPLX, self).__init__(
            *args,
            use_face_contour=use_face_contour,
            use_pca=use_pca,
            flat_hand_mean=flat_hand_mean,
            **kwargs)
        self.logger = get_logger(logger)
        self.keypoint_convention = keypoint_convention
        # override the default SMPL joint regressor if available
        if joints_regressor is not None:
            joints_regressor = torch.tensor(
                np.load(joints_regressor), dtype=torch.float)
            self.register_buffer('joints_regressor', joints_regressor)

        # allow for extra joints to be regressed if available
        if extra_joints_regressor is not None:
            joints_regressor_extra = torch.tensor(
                np.load(extra_joints_regressor), dtype=torch.float)
            self.register_buffer('joints_regressor_extra',
                                 joints_regressor_extra)

    def forward(self,
                *args,
                return_verts: bool = True,
                return_full_pose: bool = False,
                **kwargs) -> dict:
        """Forward function.

        Args:
            *args: extra arguments for SMPLX
            return_verts: whether to return vertices
            return_full_pose: whether to return full pose parameters
            **kwargs: extra arguments for SMPLX

        Returns:
            output: contains output parameters and attributes
        """

        kwargs['get_skin'] = True
        smplx_output = super(SMPLX, self).forward(*args, **kwargs)

        if not hasattr(self, 'joints_regressor'):
            joints = smplx_output.joints
        else:
            joints = vertices2joints(self.joints_regressor,
                                     smplx_output.vertices)

        if hasattr(self, 'joints_regressor_extra'):
            extra_joints = vertices2joints(self.joints_regressor_extra,
                                           smplx_output.vertices)
            joints = torch.cat([joints, extra_joints], dim=1)

        output = dict(
            global_orient=smplx_output.global_orient,
            body_pose=smplx_output.body_pose,
            joints=joints,
            betas=smplx_output.betas)

        if return_verts:
            output['vertices'] = smplx_output.vertices
        if return_full_pose:
            output['full_pose'] = smplx_output.full_pose
        return output

    def get_joint_number(self) -> int:
        """Get the number of joint from this body_model.

        Returns:
            int: Number of joint(keypoint).
        """
        return len(get_keypoints_factory()[self.keypoint_convention])



# yapf: disable
import numpy as np
from typing import Union

from xrmocap.ops.triangulation.builder import (
    BaseTriangulator, build_triangulator,
)
from xrmocap.utils.fourdag_utils import LimbInfo

# yapf: enable


class FourDAGBaseOptimizer():

    def __init__(self,
                 triangulator: Union[None, dict, BaseTriangulator] = None,
                 kps_convention='fourdag_19',
                 min_triangulate_cnt: int = 15,
                 triangulate_thresh: float = 0.05,
                 logger=None):
        """Base class for fourdag optimizater.

        Args:
            triangulator:
                triangulator to construct 3D keypoints
            kps_convention (str):
                The name of keypoints convention.
            min_triangulate_cnt (int):
                the minimum amount of 3D keypoints to be accepted
            triangulate_thresh (float):
                the maximal triangulate loss to be accepted
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        if isinstance(triangulator, dict):
            self.triangulator = build_triangulator(triangulator)
        else:
            self.triangulator = triangulator

        self.kps_convention = kps_convention
        self.min_triangulate_cnt = min_triangulate_cnt
        self.triangulate_thresh = triangulate_thresh

        self.projs = None
        self.trace_limbs = dict()
        self.trace_limb_infos = dict()
        self.limb_info = LimbInfo(self.kps_convention)

    def triangulate_person(self, limb2d):
        kps2d = limb2d.T.reshape((-1, self.limb_info.get_kps_number(), 3))
        kps3d = self.triangulator.triangulate(kps2d)
        mask = self.triangulator.loss < self.triangulate_thresh
        limb = np.concatenate((kps3d.T, mask.reshape(1, -1)), axis=0)
        return limb

    def set_cameras(self, camera_parameters):
        self.triangulator.set_cameras(camera_parameters)
        self.projs = self.triangulator.projs

    def update(self, limbs2d):
        for pidx, corr_id in enumerate(limbs2d):
            limb = self.triangulate_person(limbs2d[corr_id])
            active = sum(limb[3] > 0) >= self.min_triangulate_cnt
            if corr_id in self.trace_limbs:
                if active:
                    self.trace_limbs[corr_id] = limb
                else:
                    self.trace_limbs.pop(corr_id)
            elif active:
                self.trace_limbs[corr_id] = limb

        return self.trace_limbs



import cv2
import logging
import numpy as np
from tqdm import tqdm
from typing import List, Tuple, Union
from xrprimer.utils.ffmpeg_utils import video_to_array
from xrprimer.utils.log_utils import get_logger

from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.transform.convention.keypoints_convention import get_keypoint_num

try:
    from mmcv import digit_version
    from mmpose import __version__ as mmpose_version
    from mmpose.apis import inference_top_down_pose_model, init_pose_model
    has_mmpose = True
    import_exception = ''
except (ImportError, ModuleNotFoundError):
    has_mmpose = False
    import traceback
    stack_str = ''
    for line in traceback.format_stack():
        if 'frozen' not in line:
            stack_str += line + '\n'
    import_exception = traceback.format_exc() + '\n'
    import_exception = stack_str + import_exception


class MMposeTopDownEstimator:

    def __init__(self,
                 mmpose_kwargs: dict,
                 bbox_thr: float = 0.0,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Init a detector from mmpose.

        Args:
            mmpose_kwargs (dict):
                A dict contains args of mmpose.apis.init_detector.
                Necessary keys: config, checkpoint
                Optional keys: device
            bbox_thr (float, optional):
                Threshold of a bbox. Those have lower scores will be ignored.
                Defaults to 0.0.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        # build the pose model from a config file and a checkpoint file
        self.pose_model = init_pose_model(**mmpose_kwargs)
        # mmpose inference api takes one image per call
        self.batch_size = 1
        self.bbox_thr = bbox_thr
        self.logger = get_logger(logger)
        if not has_mmpose:
            self.logger.error(import_exception)
            raise ModuleNotFoundError(
                'Please install mmpose to run detection.')
        self.use_old_api = False
        if digit_version(mmpose_version) <= digit_version('0.13.0'):
            self.use_old_api = True

    def get_keypoints_convention_name(self) -> str:
        """Get data_source from dataset type in config file of the pose model.

        Returns:
            str:
                Name of the keypoints convention. Must be
                a key of KEYPOINTS_FACTORY.
        """
        return __translate_data_source__(
            self.pose_model.cfg.data['test']['type'])

    def infer_array(self,
                    image_array: Union[np.ndarray, list],
                    bbox_list: Union[tuple, list],
                    disable_tqdm: bool = False,
                    return_heatmap: bool = False) -> Tuple[list, list]:
        """Infer frames already in memory(ndarray type).

        Args:
            image_array (Union[np.ndarray, list]):
                BGR image ndarray in shape [n_frame, height, width, 3],
                or a list of image ndarrays in shape [height, width, 3] while
                len(list) == n_frame.
            bbox_list (Union[tuple, list]):
                A list of human bboxes.
                Shape of the nested lists is (n_frame, n_human, 5).
                Each bbox is a bbox_xyxy with a bbox_score at last.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            return_heatmap (bool, optional):
                Whether to return heatmap.
                Defaults to False.

        Returns:
            Tuple[list, list]:
                keypoints_list (list):
                    A list of human keypoints.
                    Shape of the nested lists is
                    (n_frame, n_human, n_keypoints, 3).
                    Each keypoint is an array of (x, y, confidence).
                heatmap_list (list):
                    A list of keypoint heatmaps. len(heatmap_list) == n_frame
                    and the shape of heatmap_list[f] is
                    (n_human, n_keypoints, width, height).
                bbox_list (list):
                    A list of human bboxes.
                    Shape of the nested lists is (n_frame, n_human, 5).
                    Each bbox is a bbox_xyxy with a bbox_score at last.
                    It could be smaller than the input bbox_list,
                    if there's no keypoints detected in some bbox.
        """
        ret_kps_list = []
        ret_heatmap_list = []
        ret_bbox_list = []
        n_frame = len(image_array)
        n_kps = get_keypoint_num(self.get_keypoints_convention_name())
        for start_index in tqdm(
                range(0, n_frame, self.batch_size), disable=disable_tqdm):
            end_index = min(n_frame, start_index + self.batch_size)
            # mmpose takes only one frame
            img_arr = image_array[start_index]
            person_results = []
            for frame_index in range(start_index, end_index, 1):
                bboxes_in_frame = []
                for idx, bbox in enumerate(bbox_list[frame_index]):
                    if bbox[4] > 0.0:
                        bboxes_in_frame.append({'bbox': bbox, 'id': idx})
                person_results = bboxes_in_frame
            if not self.use_old_api:
                img_input = dict(imgs_or_paths=img_arr)
            else:
                img_input = dict(img_or_path=img_arr)
            if len(bboxes_in_frame) > 0:
                pose_results, returned_outputs = inference_top_down_pose_model(
                    model=self.pose_model,
                    person_results=person_results,
                    bbox_thr=self.bbox_thr,
                    format='xyxy',
                    dataset=self.pose_model.cfg.data['test']['type'],
                    return_heatmap=return_heatmap,
                    outputs=None,
                    **img_input)
                frame_kps_results = np.zeros(
                    shape=(
                        len(bbox_list[frame_index]),
                        n_kps,
                        3,
                    ))
                frame_heatmap_results = [
                    None,
                ] * len(bbox_list[frame_index])
                frame_bbox_results = np.zeros(
                    shape=(len(bbox_list[frame_index]), 5))
                for idx, person_dict in enumerate(pose_results):
                    id = person_dict['id']
                    bbox = person_dict['bbox']
                    keypoints = person_dict['keypoints']
                    frame_bbox_results[id] = bbox
                    frame_kps_results[id] = keypoints
                    if return_heatmap:
                        # returned_outputs[0]['heatmap'].shape:
                        # 1, 133, 96, 72
                        frame_heatmap_results[id] = returned_outputs[0][
                            'heatmap'][idx]
                        ret_heatmap_list += [frame_heatmap_results]
                frame_kps_results = frame_kps_results.tolist()
                frame_bbox_results = frame_bbox_results.tolist()
            else:
                frame_kps_results = []
                frame_bbox_results = []
            ret_kps_list += [frame_kps_results]
            ret_bbox_list += [frame_bbox_results]
        return ret_kps_list, ret_heatmap_list, ret_bbox_list

    def infer_frames(
            self,
            frame_path_list: list,
            bbox_list: Union[tuple, list],
            disable_tqdm: bool = False,
            return_heatmap: bool = False,
            load_batch_size: Union[None, int] = None) -> Tuple[list, list]:
        """Infer frames from file.

        Args:
            frame_path_list (list):
                A list of frames' absolute paths.
            bbox_list (Union[tuple, list]):
                A list of human bboxes.
                Shape of the nested lists is (n_frame, n_human, 5).
                Each bbox is a bbox_xyxy with a bbox_score at last.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            return_heatmap (bool, optional):
                Whether to return heatmap.
                Defaults to False.
            load_batch_size (Union[None, int], optional):
                How many frames are loaded at the same time.
                Defaults to None, load all frames in frame_path_list.

        Returns:
            Tuple[list, list]:
                keypoints_list (list):
                    A list of human keypoints.
                    Shape of the nested lists is
                    (n_frame, n_human, n_keypoints, 3).
                    Each keypoint is an array of (x, y, confidence).
                heatmap_list (list):
                    A list of keypoint heatmaps. len(heatmap_list) == n_frame
                    and the shape of heatmap_list[f] is
                    (n_human, n_keypoints, width, height).
        """
        ret_kps_list = []
        ret_heatmap_list = []
        ret_boox_list = []
        if load_batch_size is None:
            load_batch_size = len(frame_path_list)
        for start_idx in range(0, len(frame_path_list), load_batch_size):
            end_idx = min(len(frame_path_list), start_idx + load_batch_size)
            if load_batch_size < len(frame_path_list):
                self.logger.info(
                    'Processing mmpose on frames' +
                    f'({start_idx}-{end_idx})/{len(frame_path_list)}')
            image_array_list = []
            for frame_abs_path in frame_path_list[start_idx:end_idx]:
                img_np = cv2.imread(frame_abs_path)
                image_array_list.append(img_np)
            batch_pose_list, batch_heatmap_list, batch_boox_list = \
                self.infer_array(
                    image_array=image_array_list,
                    bbox_list=bbox_list[start_idx:end_idx],
                    disable_tqdm=disable_tqdm,
                    return_heatmap=return_heatmap)
            ret_kps_list += batch_pose_list
            ret_heatmap_list += batch_heatmap_list
            ret_boox_list += batch_boox_list
        return ret_kps_list, ret_heatmap_list, ret_boox_list

    def infer_video(self,
                    video_path: str,
                    bbox_list: Union[tuple, list],
                    disable_tqdm: bool = False,
                    return_heatmap: bool = False) -> Tuple[list, list]:
        """Infer frames from a video file.

        Args:
            video_path (str):
                Path to the video to be detected.
            bbox_list (Union[tuple, list]):
                A list of human bboxes.
                Shape of the nested lists is (n_frame, n_human, 5).
                Each bbox is a bbox_xyxy with a bbox_score at last.
            disable_tqdm (bool, optional):
                Whether to disable the entire progressbar wrapper.
                Defaults to False.
            return_heatmap (bool, optional):
                Whether to return heatmap.
                Defaults to False.

        Returns:
            Tuple[list, list]:
                keypoints_list (list):
                    A list of human keypoints.
                    Shape of the nested lists is
                    (n_frame, n_human, n_keypoints, 3).
                    Each keypoint is an array of (x, y, confidence).
                heatmap_list (list):
                    A list of keypoint heatmaps. len(heatmap_list) == n_frame
                    and the shape of heatmap_list[f] is
                    (n_human, n_keypoints, width, height).
        """
        image_array = video_to_array(input_path=video_path, logger=self.logger)
        ret_kps_list, ret_heatmap_list, ret_boox_list = self.infer_array(
            image_array=image_array,
            bbox_list=bbox_list,
            disable_tqdm=disable_tqdm,
            return_heatmap=return_heatmap)
        return ret_kps_list, ret_heatmap_list, ret_boox_list

    def get_keypoints_from_result(
            self, kps2d_list: List[list]) -> Union[Keypoints, None]:
        """Convert returned keypoints2d into an instance of class Keypoints.

        Args:
            kps2d_list (List[list]):
                A list of human keypoints, returned by
                infer methods.
                Shape of the nested lists is
                (n_frame, n_human, n_keypoints, 3).

        Returns:
            Union[Keypoints, None]:
                An instance of Keypoints with mask and
                convention, data type is numpy.
                If no one has been detected in any frame,
                a None will be returned.
        """
        # shape: (n_frame, n_human, n_keypoints, 3)
        n_frame = len(kps2d_list)
        human_count_list = [len(human_list) for human_list in kps2d_list]
        if len(human_count_list) > 0:
            n_human = max(human_count_list)
        else:
            n_human = 0
        n_keypoints = get_keypoint_num(self.get_keypoints_convention_name())
        if n_human > 0:
            kps2d_arr = np.zeros(shape=(n_frame, n_human, n_keypoints, 3))
            mask_arr = np.ones_like(kps2d_arr[..., 0], dtype=np.uint8)
            for f_idx in range(n_frame):
                if len(kps2d_list[f_idx]) <= 0:
                    mask_arr[f_idx, ...] = 0
                    continue
                for h_idx in range(n_human):
                    if h_idx < len(kps2d_list[f_idx]):
                        mask_arr[f_idx, h_idx, ...] = np.sign(
                            np.array(kps2d_list[f_idx][h_idx])[:, -1])
                        kps2d_arr[f_idx,
                                  h_idx, :, :] = kps2d_list[f_idx][h_idx]
                    else:
                        mask_arr[f_idx, h_idx, ...] = 0
            keypoints2d = Keypoints(
                kps=kps2d_arr,
                mask=mask_arr,
                convention=self.get_keypoints_convention_name(),
                logger=self.logger)
        else:
            keypoints2d = None
        return keypoints2d


def __translate_data_source__(mmpose_dataset_name):
    if mmpose_dataset_name == 'TopDownSenseWholeBodyDataset':
        return 'sense_whole_body'
    elif mmpose_dataset_name == 'TopDownCocoWholeBodyDataset':
        return 'coco_wholebody'
    else:
        raise NotImplementedError



# yapf: disable
import copy
import numpy as np
from typing import Union

from xrmocap.ops.triangulation.builder import BaseTriangulator
from xrmocap.transform.keypoints3d.optim.fourdag_base_optimizer import (
    FourDAGBaseOptimizer,
)
from xrmocap.utils.fourdag_utils import (
    LimbInfo, rodrigues, rodrigues_jacobi, welsch,
)

# yapf: enable


class PersonInfo():

    def __init__(self, kps_convention) -> None:
        """save limb information."""
        self.kps_convention = kps_convention
        self.limb_info = LimbInfo(self.kps_convention)
        self.boneLen = np.zeros(
            self.limb_info.get_kps_number() - 1, dtype=np.float32)
        self.boneCnt = np.zeros(
            self.limb_info.get_kps_number() - 1, dtype=np.float32)
        self.active = 0.0
        self.shape_fixed = False
        self.data = np.zeros(
            3 + self.limb_info.get_kps_number() * 3 +
            self.limb_info.get_shape_size(),
            dtype=np.float32)

    def push_previous_bones(self, limb):
        for kps_id in range(1, self.limb_info.get_kps_number()):
            prt_idx = self.limb_info.get_kps_parent()[kps_id]
            if limb[3, kps_id] > 0 and limb[3, prt_idx] > 0:
                len = np.linalg.norm(limb[:, kps_id][:3] -
                                     limb[:, prt_idx][:3])
                self.boneLen[kps_id - 1] = (self.boneCnt[kps_id - 1] *
                                            self.boneLen[kps_id - 1] + len) / (
                                                self.boneCnt[kps_id - 1] + 1)
                self.boneCnt[kps_id - 1] += 1

    def get_trans(self):
        return self.data[:3]

    def get_pose(self):
        return self.data[3:3 + self.limb_info.get_kps_number() * 3]

    def get_trans_pose(self):
        return self.data[:3 + self.limb_info.get_kps_number() * 3]

    def get_shape(self):
        return self.data[-self.limb_info.get_shape_size():]


class SolverTerm():

    def __init__(self,
                 w_kps3d=0,
                 w_bone3d=0,
                 w_kps2d=0,
                 projs=None,
                 w_temporal_trans=0,
                 w_temporal_pose=0,
                 w_temporal_shape=0,
                 w_regular_pose=0,
                 w_regular_shape=0,
                 w_square_shape=0):
        """save some weight and information when conducting optimization."""
        # kps 3d
        self.w_kps3d = w_kps3d
        self.kps3d_target = None

        # bone 3d
        self.w_bone3d = w_bone3d
        self.bone3d_target = None

        # kps 2d
        self.w_kps2d = w_kps2d
        self.projs = projs
        self.kps2d_target = None

        # temporal
        self.w_temporal_trans = w_temporal_trans
        self.w_temporal_pose = w_temporal_pose
        self.w_temporal_shape = w_temporal_shape
        self.paramPrev = None

        # regular
        self.w_regular_pose = w_regular_pose
        self.w_regular_shape = w_regular_shape
        self.w_square_shape = w_square_shape

    def set_kps3d_target(self, kps3d_target):
        self.kps3d_target = kps3d_target

    def set_bone3d_target(self, bone3d_target):
        self.bone3d_target = bone3d_target

    def set_kps2d_target(self, kps2d_target):
        self.kps2d_target = kps2d_target

    def set_paramPrev(self, paramPrev):
        self.paramPrev = paramPrev


class LimbSolver():

    def __init__(self, kps_convention) -> None:
        """slove human pose and shape."""
        self.kps_convention = kps_convention
        self.limb_info = LimbInfo(self.kps_convention)
        self.m_kps = np.array(self.limb_info.get_kps_prior()).reshape(
            3, self.limb_info.get_kps_number())
        self.shape_blend = np.array(self.limb_info.get_shape_blend()).reshape(
            self.limb_info.get_kps_number() * 3,
            self.limb_info.get_shape_size())
        self.bone_shape_blend = np.zeros(
            (3 * (self.limb_info.get_kps_number() - 1),
             self.limb_info.get_shape_size()),
            dtype=np.float32)
        self.limb_info = LimbInfo(self.kps_convention)
        for kps_id in range(1, self.limb_info.get_kps_number()):
            self.bone_shape_blend[3 * (kps_id - 1):3 * (kps_id - 1)+3]\
                    = self.shape_blend[3 * kps_id:3 * kps_id+3] \
                    - self.shape_blend[
                        3 * self.limb_info.get_kps_parent()[kps_id]:
                            3 * self.limb_info.get_kps_parent()[kps_id]+3]

    def cal_kps_with_chain_warps(self, chain_warps):
        kps_final = np.zeros((3, int(chain_warps.shape[1] / 4)),
                             dtype=np.float32)
        for kps_id in range(kps_final.shape[1]):
            kps_final[:, kps_id] = (chain_warps[0:0 + 3, 4 * kps_id +
                                                3:4 * kps_id + 3 + 1]).reshape(
                                                    (-1))
        return kps_final

    def cal_kps_with_param(self, param, j_cut=-1):
        j_cut = j_cut if j_cut > 0 else self.m_kps.shape[1]
        kps_blend = self.cal_kps_blend(param)
        return self.cal_kps_with_chain_warps(
            self.cal_chain_warps(
                self.cal_node_warps(param, kps_blend[:, :j_cut])))

    def cal_kps_blend(self, param):
        kps_offset = np.matmul(self.shape_blend, param.get_shape())
        kps_blend = self.m_kps + kps_offset.reshape((self.m_kps.shape[1], 3)).T
        return kps_blend

    def cal_node_warps(self, param, kps_blend):
        node_warps = np.zeros((4, kps_blend.shape[1] * 4), dtype=np.float32)
        for kps_id in range(kps_blend.shape[1]):
            matrix = np.identity(4, dtype=np.float32)
            if kps_id == 0:
                matrix[:3, -1:] = (kps_blend[:, kps_id] +
                                   param.get_trans()).reshape((-1, 1))
            else:
                matrix[:3, -1:] = (
                    kps_blend[:, kps_id] -
                    kps_blend[:, self.limb_info.get_kps_parent()[kps_id]]
                ).reshape((-1, 1))

            matrix[:3, :3] = rodrigues(param.get_pose()[3 * kps_id:3 * kps_id +
                                                        3])
            node_warps[:4, 4 * kps_id:4 * kps_id + 4] = matrix
        return node_warps

    def cal_chain_warps(self, node_warps):
        chain_warps = np.zeros((4, node_warps.shape[1]), dtype=np.float32)
        for kps_id in range(int(node_warps.shape[1] / 4)):
            if kps_id == 0:
                chain_warps[:, kps_id * 4:kps_id * 4 +
                            4] = node_warps[:, kps_id * 4:kps_id * 4 + 4]
            else:
                chain_warps[:, kps_id * 4:kps_id * 4 + 4] = np.matmul(
                    chain_warps[:,
                                self.limb_info.get_kps_parent()[kps_id] *
                                4:self.limb_info.get_kps_parent()[kps_id] * 4 +
                                4], node_warps[:, kps_id * 4:kps_id * 4 + 4])
        return chain_warps

    def align_root_affine(self, term, param):
        # align root affine
        param.data[:3] = term.kps3d_target[:, 0][:3] - self.m_kps[:, 0]

        def cal_axes(x_axis, y_axis):
            axes = np.zeros((3, 3), dtype=np.float32)
            axes[:, 0] = x_axis / np.linalg.norm(x_axis)
            axes[:, 2] = np.cross(x_axis, y_axis) / np.linalg.norm(
                np.cross(x_axis, y_axis))
            axes[:, 1] = np.cross(axes[:, 2], axes[:, 0]) / np.linalg.norm(
                np.cross(axes[:, 2], axes[:, 0]))
            return axes

        mat = np.matmul(
            cal_axes(term.kps3d_target[:, 2][:3] - term.kps3d_target[:, 1][:3],
                     term.kps3d_target[:, 3][:3] -
                     term.kps3d_target[:, 1][:3]), (np.linalg.inv(
                         cal_axes(self.m_kps[:, 2] - self.m_kps[:, 1],
                                  self.m_kps[:, 3] - self.m_kps[:, 1]))))
        angle = np.arccos((mat[0, 0] + mat[1, 1] + mat[2, 2] - 1) / 2)
        x = (mat[2, 1] - mat[1, 2]) / np.sqrt((mat[2, 1] - mat[1, 2])**2 +
                                              (mat[0, 2] - mat[2, 0])**2 +
                                              (mat[1, 0] - mat[0, 1])**2)
        y = (mat[0, 2] - mat[2, 0]) / np.sqrt((mat[2, 1] - mat[1, 2])**2 +
                                              (mat[0, 2] - mat[2, 0])**2 +
                                              (mat[1, 0] - mat[0, 1])**2)
        z = (mat[1, 0] - mat[0, 1]) / np.sqrt((mat[2, 1] - mat[1, 2])**2 +
                                              (mat[0, 2] - mat[2, 0])**2 +
                                              (mat[1, 0] - mat[0, 1])**2)
        param.data[3:3 + self.limb_info.get_kps_number() *
                   3][:3] = angle * np.array([x, y, z], dtype=np.float32)

    def solve_pose(self,
                   term,
                   param,
                   maxIter_time,
                   hierarchy=False,
                   update_thresh=1e-4):
        kps_blend = self.cal_kps_blend(param)
        hier_size = max(self.limb_info.get_hierarchy_map())
        hier = 0 if hierarchy else hier_size
        j_cut = 0
        while hier <= hier_size:
            while j_cut < self.limb_info.get_kps_number(
            ) and self.limb_info.get_hierarchy_map()[j_cut] <= hier:
                j_cut += 1
            for iter_time in range(maxIter_time):
                node_warps = self.cal_node_warps(param, kps_blend[:, :j_cut])
                chain_warps = self.cal_chain_warps(node_warps)
                kps_final = self.cal_kps_with_chain_warps(chain_warps)
                kps_jacobi = np.zeros((3 * j_cut, 3 + 3 * j_cut),
                                      dtype=np.float32)
                ATA = np.zeros((3 + 3 * j_cut, 3 + 3 * j_cut),
                               dtype=np.float32)
                ATb = np.zeros((3 + 3 * j_cut), dtype=np.float32)
                node_warps_jacobi = np.zeros((9, 3 * j_cut), dtype=np.float32)
                for kps_id in range(j_cut):
                    node_warps_jacobi[:, 3 * kps_id:3 * kps_id +
                                      3] = rodrigues_jacobi(
                                          param.get_pose()[3 *
                                                           kps_id:3 * kps_id +
                                                           3]).T
                for d_jidx in range(j_cut):
                    kps_jacobi[3 * d_jidx:3 * d_jidx + 3, :3] = np.identity(
                        3, dtype=np.float32)
                    for dAxis in range(3):
                        d_chain_warps = np.zeros((4, 4 * j_cut),
                                                 dtype=np.float32)
                        valid = np.zeros(j_cut, dtype=np.float32)
                        valid[d_jidx] = 1
                        d_chain_warps[:3, 4 * d_jidx:4 * d_jidx +
                                      3] = node_warps_jacobi[:, 3 * d_jidx +
                                                             dAxis].copy(
                                                             ).reshape(
                                                                 (3, 3)).T
                        if d_jidx != 0:
                            d_chain_warps[:, 4 * d_jidx:4 * d_jidx +
                                          4] = np.matmul(
                                              chain_warps[:,
                                                          4 * self.limb_info.
                                                          get_kps_parent(
                                                          )[d_jidx]:4 *
                                                          self.limb_info.
                                                          get_kps_parent(
                                                          )[d_jidx] + 4],
                                              d_chain_warps[:, 4 *
                                                            d_jidx:4 * d_jidx +
                                                            4])

                        for kps_id in range(d_jidx + 1, j_cut):
                            prt_idx = self.limb_info.get_kps_parent()[kps_id]
                            valid[kps_id] = valid[prt_idx]
                            if valid[kps_id]:
                                d_chain_warps[:, 4 * kps_id:4 * kps_id +
                                              4] = np.matmul(
                                                  d_chain_warps[:,
                                                                4 * prt_idx:4 *
                                                                prt_idx + 4],
                                                  node_warps[:, 4 * kps_id:4 *
                                                             kps_id + 4])
                                kps_jacobi[kps_id * 3:kps_id * 3 + 3,
                                           3 + d_jidx * 3 + dAxis:3 +
                                           d_jidx * 3 + dAxis +
                                           1] = d_chain_warps[0:0 + 3,
                                                              4 * kps_id +
                                                              3:4 * kps_id +
                                                              3 + 1]
                if term.w_kps3d > 0:
                    for kps_id in range(j_cut):
                        if term.kps3d_target[3, kps_id] > 0:
                            w = term.w_kps3d * term.kps3d_target[3, kps_id]
                            jacobi = kps_jacobi[3 * kps_id:3 * kps_id + 3]
                            ATA += w * np.matmul(jacobi.T, jacobi)
                            ATb += w * np.matmul(
                                jacobi.T,
                                (term.kps3d_target[0:0 + 3, kps_id:kps_id + 1]
                                 - kps_final[:, kps_id].reshape(
                                     (-1, 1)))).reshape(-1)

                if term.w_kps2d > 0:
                    for view in range(int(term.projs.shape[1] / 4)):
                        kps2d_target = term.kps2d_target[:,
                                                         view * self.limb_info.
                                                         get_kps_number():
                                                         view * self.limb_info.
                                                         get_kps_number() +
                                                         self.limb_info.
                                                         get_kps_number()]
                        if sum(kps2d_target[2] > 0) > 0:
                            proj = term.projs[:, view * 4:view * 4 + 4]
                            for kps_id in range(j_cut):
                                if kps2d_target[2, kps_id] > 0:
                                    abc = np.matmul(
                                        proj,
                                        np.append(kps_final[:, kps_id], 1))
                                    proj_jacobi = np.array(
                                        [
                                            1.0 / abc[2], 0.0, -abc[0] /
                                            (abc[2] * abc[2]), 0.0, 1.0 /
                                            abc[2], -abc[1] / (abc[2] * abc[2])
                                        ],
                                        dtype=np.float32).reshape((2, 3))
                                    proj_jacobi = np.matmul(
                                        proj_jacobi, proj[:, :3])

                                    w = term.w_kps2d * kps2d_target[2, kps_id]
                                    jacobi = np.matmul(
                                        proj_jacobi,
                                        kps_jacobi[3 * kps_id:3 * kps_id + 3])

                                    ATA += w * np.matmul(jacobi.T, jacobi)
                                    ATb += w * np.matmul(
                                        jacobi.T,
                                        kps2d_target[:2, kps_id:kps_id +
                                                     1].reshape(-1) -
                                        abc[:2] / abc[2])

                if term.w_temporal_trans > 0:
                    ATA[:3, :3] += term.w_temporal_trans * np.identity(
                        3, dtype=np.float32)
                    ATb[:3] += term.w_temporal_trans * (
                        term.paramPrev.get_trans() - param.get_trans())

                if term.w_temporal_pose > 0:
                    ATA[-3 * j_cut:,
                        -3 * j_cut:] += term.w_temporal_pose * np.identity(
                            3 * j_cut, dtype=np.float32)
                    ATb[-3 * j_cut:] += term.w_temporal_pose * (
                        term.paramPrev.get_pose()[:3 * j_cut] -
                        param.get_pose()[:3 * j_cut])

                if term.w_regular_pose > 0:
                    ATA += term.w_regular_pose * np.identity(
                        3 + 3 * j_cut, dtype=np.float32)

                delta = np.linalg.solve(ATA, ATb)
                param.data[:3 + self.limb_info.get_kps_number() *
                           3][:3 + 3 * j_cut] += delta

                if np.linalg.norm(delta) < update_thresh:
                    break
            hier += 1

    def solve_shape(self, term, param, maxIter_time, update_thresh=1e-4):
        for iter_time in range(maxIter_time):
            # calc status
            kps_blend = self.cal_kps_blend(param)
            ATA = np.zeros((self.limb_info.get_shape_size(),
                            self.limb_info.get_shape_size()),
                           dtype=np.float32)
            ATb = np.zeros(self.limb_info.get_shape_size(), dtype=np.float32)

            if term.w_bone3d > 0:
                for kps_id in range(1, self.limb_info.get_kps_number()):
                    if term.bone3d_target[1, kps_id - 1] > 0:
                        w = term.w_bone3d * term.bone3d_target[1, kps_id - 1]
                        prt_idx = self.limb_info.get_kps_parent()[kps_id]
                        dir = kps_blend[:, kps_id] - kps_blend[:, prt_idx]
                        jacobi = self.bone_shape_blend[3 * (kps_id - 1):3 *
                                                       (kps_id - 1) + 3]
                        ATA += w * np.matmul(jacobi.T, jacobi)
                        ATb += w * np.matmul(
                            jacobi.T, term.bone3d_target[0, kps_id - 1] *
                            (dir / np.linalg.norm(dir)) - dir)

            if term.w_kps3d > 0 or term.w_kps2d > 0:
                chain_warps = self.cal_chain_warps(
                    self.cal_node_warps(param, kps_blend))
                kps_final = self.cal_kps_with_chain_warps(chain_warps)
                kps_jacobi = np.zeros((3 * self.limb_info.get_kps_number(),
                                       self.limb_info.get_shape_size()),
                                      dtype=np.float32)
                for kps_id in range(self.limb_info.get_kps_number()):
                    if kps_id == 0:
                        kps_jacobi[3 * kps_id:3 * kps_id +
                                   3] = self.shape_blend[3 *
                                                         kps_id:3 * kps_id + 3]
                    else:
                        prt_idx = self.limb_info.get_kps_parent()[kps_id]
                        kps_jacobi[3 * kps_id:3 * kps_id+3] =\
                            kps_jacobi[3 * prt_idx:3 * prt_idx+3] \
                            + chain_warps[:3, 4*prt_idx+3] * (
                                self.shape_blend[3 * kps_id:3 * kps_id+3] -
                                self.shape_blend[3 * prt_idx:3 * prt_idx+3])

                if term.w_kps3d > 0:
                    for kps_id in range(self.limb_info.get_kps_number()):
                        if term.kps3d_target[3, kps_id] > 0:
                            w = term.w_kps3d * term.kps3d_target[3, kps_id]
                            jacobi = kps_jacobi[3 * kps_id:3 * kps_id + 3]
                            ATA += w * np.matmul(jacobi.T, jacobi)
                            ATb += w * np.matmul(
                                jacobi.T,
                                (term.kps3d_target[0 + 3, kps_id + 1] -
                                 kps_final[:, kps_id]))

                if term.w_kps2d > 0:
                    for view in range(int(len(term.projs[0]) / 4)):
                        kps2d_target = term.kps2d_target[
                            view * self.limb_info.get_kps_number():view *
                            self.limb_info.get_kps_number() +
                            self.limb_info.get_kps_number()]
                        if sum(kps2d_target[2] > 0) > 0:
                            proj = term.projs[view * 4:view * 4 + 4]
                            for kps_id in range(
                                    self.limb_info.get_kps_number()):
                                if kps2d_target[2, kps_id] > 0:
                                    abc = proj * np.append(
                                        kps_final[:, kps_id], 1)
                                    proj_jacobi = np.zeros((2, 3),
                                                           dtype=np.float32)
                                    proj_jacobi = np.array(
                                        [
                                            1.0 / abc[2], 0.0, -abc[0] /
                                            (abc[2] * abc[2]), 0.0, 1.0 /
                                            abc[2], -abc[1] / (abc[2] * abc[2])
                                        ],
                                        dtype=np.float32).reshape((2, 3))
                                    proj_jacobi = proj_jacobi * proj[:, :3]

                                    w = term.w_kps2d * kps2d_target[2, kps_id]
                                    jacobi = proj_jacobi * kps_jacobi[
                                        3 * kps_id:3 * kps_id + 3]
                                    ATA += w * np.matmul(jacobi.T, jacobi)
                                    ATb += w * np.matmul(
                                        jacobi.T,
                                        kps2d_target[0 + 2, kps_id + 1] -
                                        abc[:2] / abc[2])

            if term.w_temporal_shape > 0:
                ATA += term.w_temporal_shape * np.identity(
                    self.limb_info.get_shape_size(), dtype=np.float32)
                ATb += term.w_temporal_shape * (
                    term.paramPrev.get_shape() - param.get_shape())

            if term.w_square_shape > 0:
                ATA += term.w_square_shape * np.identity(
                    self.limb_info.get_shape_size(), dtype=np.float32)
                ATb -= term.w_square_shape * param.get_shape()

            if term.w_regular_shape > 0:
                ATA += term.w_regular_shape * np.identity(
                    self.limb_info.get_shape_size(), dtype=np.float32)

            delta = np.linalg.solve(ATA, ATb)
            param.data[-self.limb_info.get_shape_size():] += delta

            if np.linalg.norm(delta) < update_thresh:
                break


class FourDAGOptimizer(FourDAGBaseOptimizer):

    def __init__(self,
                 triangulator: Union[None, dict, BaseTriangulator] = None,
                 active_rate: float = 0.1,
                 min_track_cnt: int = 5,
                 bone_capacity: int = 100,
                 w_bone3d: float = 1.0,
                 w_square_shape: float = 1e-2,
                 shape_max_iter: int = 5,
                 w_kps3d: float = 1.0,
                 w_regular_pose: float = 1e-3,
                 pose_max_iter: int = 20,
                 w_kps2d: float = 1e-5,
                 w_temporal_trans: float = 1e-1,
                 w_temporal_pose: float = 1e-2,
                 min_triangulate_cnt: int = 15,
                 init_active: float = 0.9,
                 triangulate_thresh: float = 0.05,
                 kps_convention: str = 'fourdag_19',
                 logger=None):
        """optimize with 2D projections loss, shape prior and temporal
        smoothing.

        Args:
            triangulator:
                triangulator to construct 3D keypoints
            active_rate (float):
                active value degression rate
            min_track_cnt (int):
                minimum track value
            bone_capacity (int):
                the minimum bone capacity to turn to optimization
            w_bone3d (float):
                weight for 3D kepoints loss to solve shape
            w_square_shape (float):
                weight for shape regulization loss
            shape_max_iter (int):
                maximal iteration to solve shape
            w_kps3d (float):
                weight for 3D kepoints loss to solve pose
            w_regular_pose (float):
                weight for pose regulization loss
            pose_max_iter (int):
                maximal iteration to solve pose
            w_kps2d (float):
                weight for 2D kepoints loss to solve pose
            w_temporal_trans (float):
                weight for temporal smoothing loss to solve pose
            w_temporal_pose (float):
                weight for temporal smoothing loss to solve pose
            min_triangulate_cnt (int):
                the minimum amount of 3D keypoints to be accepted
            init_active:
                initial weight for active value
            triangulate_thresh (float):
                the maximal triangulate loss to be accepted
            kps_convention (str):
                The name of keypoints convention.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """

        super().__init__(
            triangulator=triangulator,
            kps_convention=kps_convention,
            min_triangulate_cnt=min_triangulate_cnt,
            triangulate_thresh=triangulate_thresh,
            logger=logger)
        self.active_rate = active_rate
        self.min_track_cnt = min_track_cnt
        self.bone_capacity = bone_capacity
        self.w_bone3d = w_bone3d
        self.w_square_shape = w_square_shape
        self.shape_max_iter = shape_max_iter
        self.w_kps3d = w_kps3d
        self.w_regular_pose = w_regular_pose
        self.pose_max_iter = pose_max_iter
        self.w_kps2d = w_kps2d
        self.w_temporal_trans = w_temporal_trans
        self.w_temporal_pose = w_temporal_pose
        self.init_active = init_active
        self.limb_solver = LimbSolver(kps_convention)
        self.limb_info = LimbInfo(self.kps_convention)

    def update(self, limbs2d):
        for pid, corr_id in enumerate(limbs2d):
            if corr_id in self.trace_limbs:
                info = self.trace_limb_infos[corr_id]
                limb = self.trace_limbs[corr_id]
                active = min(
                    info.active + self.active_rate *
                    (2.0 * welsch(self.min_track_cnt,
                                  sum(limbs2d[corr_id][2] > 0)) - 1.0), 1.0)
                if info.active < 0:
                    self.trace_limbs.pop(corr_id)
                    self.trace_limb_infos.pop(corr_id)
                    continue

                if not info.shape_fixed:
                    limb = self.triangulate_person(limbs2d[corr_id])
                    if sum(limb[3] > 0) >= self.min_triangulate_cnt:
                        info.push_previous_bones(limb)
                        if min(info.boneCnt) >= self.bone_capacity:
                            info.push_previous_bones(limb)
                            shape_term = SolverTerm(
                                w_bone3d=self.w_bone3d,
                                w_square_shape=self.w_square_shape)
                            shape_term.set_bone3d_target(
                                np.row_stack((info.boneLen.T,
                                              np.ones(
                                                  info.boneLen.shape[0],
                                                  dtype=np.float32))))
                            self.limb_solver.solve_shape(
                                shape_term, info, self.shape_max_iter)

                            # align pose
                            pose_term = SolverTerm(
                                w_kps3d=self.w_kps3d,
                                w_regular_pose=self.w_regular_pose)
                            pose_term.set_kps3d_target(limb)
                            self.limb_solver.align_root_affine(pose_term, info)
                            self.limb_solver.solve_pose(
                                pose_term, info, self.pose_max_iter)
                            limb[:3] = self.limb_solver.cal_kps_with_param(
                                info)
                            info.shape_fixed = True
                    self.trace_limbs[corr_id] = limb

                else:
                    # align pose
                    pose_term = SolverTerm(
                        w_kps2d=self.w_kps2d,
                        projs=self.projs,
                        w_regular_pose=self.w_regular_pose,
                        w_temporal_trans=self.w_temporal_trans,
                        w_temporal_pose=self.w_temporal_pose)
                    pose_term.set_kps2d_target(copy.deepcopy(limbs2d[corr_id]))
                    # filter single view correspondence
                    corr_cnt = np.zeros(
                        self.limb_info.get_kps_number(), dtype=np.float32)
                    kps_confidence = np.ones(
                        self.limb_info.get_kps_number(), dtype=np.float32)
                    for view in range(int(self.projs.shape[1] / 4)):
                        corr_cnt += (
                            (pose_term.
                             kps2d_target[:,
                                          view * self.limb_info.get_kps_number(
                                          ):self.limb_info.get_kps_number() *
                                          (view + 1)][2].T > 0).astype(np.int))
                    for kps_id in range(self.limb_info.get_kps_number()):
                        if corr_cnt[kps_id] <= 1:
                            kps_confidence[kps_id] = 0
                            for view in range(int(self.projs.shape[1] / 4)):
                                pose_term.kps2d_target[:,
                                                       view * self.limb_info.
                                                       get_kps_number() +
                                                       kps_id] = 0

                    pose_term.set_paramPrev(info)
                    self.limb_solver.solve_pose(pose_term, info,
                                                self.pose_max_iter)
                    limb[:3] = self.limb_solver.cal_kps_with_param(info)
                    limb[3] = kps_confidence.T
                    # update active
                    info.active = active
            else:
                limb = self.triangulate_person(limbs2d[corr_id])
                # alloc new person
                if sum(limb[3] > 0) >= self.min_triangulate_cnt:
                    self.trace_limb_infos[corr_id] = PersonInfo(
                        self.kps_convention)
                    info = self.trace_limb_infos[corr_id]
                    info.push_previous_bones(limb)
                    info.active = self.init_active
                    self.trace_limbs[corr_id] = limb
        return self.trace_limbs



# yapf: disable
from .camera import (
    get_all_color_kinect_parameter_from_smc,
    get_color_camera_parameter_from_smc,
    load_camera_parameters_from_zoemotion_dir,
)
from .h5py_helper import H5Helper
from .image import (
    get_n_frame_from_mview_src, load_clip_from_mview_src,
    load_multiview_images,
)

# yapf: enable

__all__ = [
    'H5Helper', 'get_all_color_kinect_parameter_from_smc',
    'get_color_camera_parameter_from_smc', 'get_n_frame_from_mview_src',
    'load_camera_parameters_from_zoemotion_dir', 'load_clip_from_mview_src',
    'load_multiview_images'
]



from .mmdet_detector import MMdetDetector, process_mmdet_results
from .mmtrack_detector import MMtrackDetector, process_mmtrack_results

__all__ = [
    'MMdetDetector', 'MMtrackDetector', 'process_mmdet_results',
    'process_mmtrack_results'
]



from typing import Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def qsort_bbox_list(bbox_list: list,
                    only_max: bool = False,
                    bbox_convention: Literal['xyxy', 'xywh'] = 'xyxy'):
    """Sort a list of bboxes, by their area in pixel(W*H).

    Args:
        input_list (list):
            A list of bboxes. Each item is a list of (x1, y1, x2, y2)
        only_max (bool, optional):
            If True, only assure the max element at first place,
            others may not be well sorted.
            If False, return a well sorted descending list.
            Defaults to False.
        bbox_convention (str, optional):
            Bbox type, xyxy or xywh. Defaults to 'xyxy'.

    Returns:
        list:
            A sorted(maybe not so well) descending list.
    """
    if len(bbox_list) <= 1:
        return bbox_list
    else:
        bigger_list = []
        less_list = []
        anchor_index = int(len(bbox_list) / 2)
        anchor_bbox = bbox_list[anchor_index]
        anchor_area = get_area_of_bbox(anchor_bbox, bbox_convention)
        for i in range(len(bbox_list)):
            if i == anchor_index:
                continue
            tmp_bbox = bbox_list[i]
            tmp_area = get_area_of_bbox(tmp_bbox, bbox_convention)
            if tmp_area >= anchor_area:
                bigger_list.append(tmp_bbox)
            else:
                less_list.append(tmp_bbox)
        if only_max:
            return qsort_bbox_list(bigger_list) + \
                [anchor_bbox, ] + less_list
        else:
            return qsort_bbox_list(bigger_list) + \
                [anchor_bbox, ] + qsort_bbox_list(less_list)


def get_area_of_bbox(
        bbox: Union[list, tuple],
        bbox_convention: Literal['xyxy', 'xywh'] = 'xyxy') -> float:
    """Get the area of a bbox_xyxy.

    Args:
        (Union[list, tuple]):
            A list of [x1, y1, x2, y2].
        bbox_convention (str, optional):
            Bbox type, xyxy or xywh. Defaults to 'xyxy'.

    Returns:
        float:
            Area of the bbox(|y2-y1|*|x2-x1|).
    """
    if bbox_convention == 'xyxy':
        return abs(bbox[2] - bbox[0]) * abs(bbox[3] - bbox[1])
    elif bbox_convention == 'xywh':
        return abs(bbox[2] * bbox[3])
    else:
        raise TypeError(f'Wrong bbox convention: {bbox_convention}')



# yapf: disable
from .aniposelib_optimizer import AniposelibOptimizer
from .base_optimizer import BaseOptimizer
from .median_smooth import MedianSmooth, median_filter_data
from .nan_interpolation import (
    NanInterpolation, count_masked_nan, interpolate_np_data,
)
from .smpl_shape_aware_optimizer import SMPLShapeAwareOptimizer
from .trajectory_optimizer import TrajectoryOptimizer

# yapf: enable

__all__ = [
    'AniposelibOptimizer', 'BaseOptimizer', 'MedianSmooth', 'NanInterpolation',
    'SMPLShapeAwareOptimizer', 'TrajectoryOptimizer', 'count_masked_nan',
    'interpolate_np_data', 'median_filter_data'
]



import logging
from typing import Union
from xrprimer.utils.log_utils import get_logger


class BaseEstimator:
    """Base Estimator."""

    def __init__(self,
                 work_dir: str,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        self.work_dir = work_dir
        self.verbose = verbose
        self.logger = get_logger(logger)

    def run(self) -> None:
        ...



import h5py
import io
import numpy as np


class H5Helper:
    """Some helper function related to h5py."""
    h5_element_type = \
        (np.int64, np.float64, str, np.float, float, np.float32, int)
    h5_list_type = \
        (list, np.ndarray)

    @staticmethod
    def save_attrs_to_h5file(h5file: h5py.File,
                             root_key: str = '/',
                             dic: dict = {}):
        for k, v in dic.items():
            h5file[root_key].attrs[k] = v
        return None

    @staticmethod
    def recursively_save_dict_contents_to_h5file(h5file: h5py.File,
                                                 root_key: str = '/',
                                                 dic: dict = {}):
        if not root_key.endswith('/'):
            root_key = root_key + '/'
        for k, v in dic.items():
            k = str(k)
            if k == 'attrs':
                if root_key not in h5file:
                    h5file.create_group(root_key)
                H5Helper.save_attrs_to_h5file(h5file, root_key, v)
                continue
            if isinstance(v, dict):
                H5Helper.recursively_save_dict_contents_to_h5file(
                    h5file, root_key + k + '/', v)
            elif isinstance(v, H5Helper.h5_element_type):
                h5file[root_key + k] = v
            elif isinstance(v, H5Helper.h5_list_type):
                try:
                    h5file[root_key + k] = v
                except TypeError:
                    v = np.array(v).astype('|S9')
                    h5file[root_key + k] = v
            else:
                raise TypeError(f'Cannot save {type(v)} type.')
        return None

    @staticmethod
    def load_dict_from_hdf5(filename):
        with h5py.File(filename, 'r') as h5file:
            return H5Helper.recursively_load_dict_contents_from_group(
                h5file, '/')

    @staticmethod
    def load_h5group_attr_to_dict(h5file: h5py.File,
                                  root_key: str = '/') -> dict:
        ret_dict = {}
        for k, v in h5file[root_key].attrs.items():
            ret_dict[k] = v
        return ret_dict

    @staticmethod
    def recursively_load_dict_contents_from_group(h5file: h5py.File,
                                                  root_key: str = '/'):
        if not root_key.endswith('/'):
            root_key = root_key + '/'
        ret_dict = {}
        if len(h5file[root_key].attrs) > 0:
            ret_dict['attrs'] = H5Helper.load_h5group_attr_to_dict(
                h5file, root_key)
        for key, item in h5file[root_key].items():
            if isinstance(item, h5py._hl.dataset.Dataset):
                ret_dict[key] = item[...]
            elif isinstance(item, h5py._hl.group.Group):
                ret_dict[key] = \
                    H5Helper.recursively_load_dict_contents_from_group(
                    h5file, root_key + key + '/')
        return ret_dict

    @staticmethod
    def h5py_to_binary(h5f):
        bio = io.BytesIO()
        with h5py.File(bio, 'w') as biof:
            for key, value in h5f.items():
                h5f.copy(
                    value,
                    biof,
                    expand_soft=True,
                    expand_external=True,
                    expand_refs=True)
        return bio



PANOPTIC_KEYPOINTS = [
    'neck_openpose',
    'nose_openpose',
    'pelvis_openpose',
    'left_shoulder_openpose',
    'left_elbow_openpose',
    'left_wrist_openpose',
    'left_hip_openpose',
    'left_knee_openpose',
    'left_ankle_openpose',
    'right_shoulder_openpose',
    'right_elbow_openpose',
    'right_wrist_openpose',
    'right_hip_openpose',
    'right_knee_openpose',
    'right_ankle_openpose',
    'left_eye_openpose',
    'left_ear_openpose',
    'right_eye_openpose',
    'right_ear_openpose',
]



from mmcv.utils import Registry

from .affinity_estimator import AppearanceAffinityEstimator
from .multi_view_pose_transformer import MviewPoseTransformer

ARCHITECTURES = Registry('architectures')

ARCHITECTURES.register_module(
    name='AppearanceAffinityEstimator', module=AppearanceAffinityEstimator)
ARCHITECTURES.register_module(
    name='MviewPoseTransformer', module=MviewPoseTransformer)


def build_architecture(cfg):
    """Build framework."""
    return ARCHITECTURES.build(cfg)



# yapf: disable
import logging
import numpy as np
import torch
from typing import Any, Union
from xrprimer.utils.log_utils import get_logger
from xrprimer.utils.path_utils import (
    Existence, check_path_existence, check_path_suffix,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# yapf: enable


class SMPLData(dict):
    BODY_POSE_LEN = 23
    DEFAULT_BODY_JOINTS_NUM = 45
    BODY_POSE_KEYS = {
        'global_orient',
        'body_pose',
    }
    FULL_POSE_KEYS = {
        'global_orient',
        'body_pose',
    }

    def __init__(self,
                 src_dict: dict = None,
                 gender: Union[Literal['female', 'male', 'neutral'],
                               None] = None,
                 fullpose: Union[np.ndarray, torch.Tensor, None] = None,
                 transl: Union[np.ndarray, torch.Tensor, None] = None,
                 betas: Union[np.ndarray, torch.Tensor, None] = None,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Construct a SMPLData instance with pre-set values. If any of gender,
        fullpose, transl, betas is provided, it will override the item in
        source_dict.

        Args:
            src_dict (dict, optional):
                A dict with items in HumanData fashion.
                Defaults to None.
            gender (Union[
                    Literal['female', 'male', 'neutral'], None], optional):
                Gender of the body model.
                Should be one among ["female", "male", "neutral"].
                Defaults to None.
            fullpose (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for fullpose, in shape [n_frame, 24, 3].
                Defaults to None, zero-tensor will be created.
            transl (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for translation, in shape [n_frame, 3].
                Defaults to None, zero-tensor will be created.
            betas (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for translation,
                in shape [n_frame, betas_dim].
                Defaults to None,
                zero-tensor in shape [n_frame, 10] will be created.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        if src_dict is not None:
            super().__init__(src_dict)
        else:
            super().__init__()
        self.n_body_joints = self.__class__.DEFAULT_BODY_JOINTS_NUM
        self.logger = get_logger(logger)
        if gender is None and 'gender' not in self:
            gender = 'neutral'
        if gender is not None:
            self.set_gender(gender)
        if fullpose is None and 'fullpose' not in self:
            fullpose_dim = self.__class__.get_fullpose_dim()
            fullpose = np.zeros(shape=(1, fullpose_dim, 3))
        if fullpose is not None:
            self.set_fullpose(fullpose)
        if transl is None and 'transl' not in self:
            transl = np.zeros(shape=(self.get_batch_size(), 3))
        if transl is not None:
            self.set_transl(transl)
        if betas is None and 'betas' not in self:
            betas = np.zeros(shape=(self.get_batch_size(), 10))
        if betas is not None:
            self.set_betas(betas)

    @classmethod
    def fromfile(cls, npz_path: str) -> 'SMPLData':
        """Construct a body model data structure from an npz file.

        Args:
            npz_path (str):
                Path to a dumped npz file.

        Returns:
            SMPLData:
                A SMPLData instance load from file.
        """
        ret_instance = cls()
        ret_instance.load(npz_path)
        return ret_instance

    @classmethod
    def get_fullpose_dim(cls) -> int:
        """Get dimension of full pose.

        Returns:
            int:
                Dim value. Full pose shall be
                in shape (frame_n, dim, 3)
        """
        global_orient_dim = 1
        ret_sum = global_orient_dim + cls.BODY_POSE_LEN
        return ret_sum

    def set_gender(
            self,
            gender: Literal['female', 'male', 'neutral'] = 'neutral') -> None:
        """Set gender.

        Args:
            gender (Literal["female", "male", "neutral"], optional):
                Gender of the body model.
                Should be one among ["female", "male", "neutral"].
                Defaults to 'neutral'.

        Raises:
            ValueError: Value of gender is not correct.
        """
        if gender in ['female', 'male', 'neutral']:
            super().__setitem__('gender', gender)
        else:
            self.logger.error('Value of gender is not correct.\n' +
                              f'gender: {gender}.')
            raise ValueError

    def set_fullpose(self, fullpose: Union[np.ndarray, torch.Tensor]) -> None:
        """Set full pose data.

        Args:
            fullpose (Union[np.ndarray, torch.Tensor]):
                Full pose in ndarray or tensor,
                in shape [batch_size, fullpose_dim, 3].
                global_orient at [:, 0, :].

        Raises:
            TypeError: Type of fullpose is not correct.
        """
        fullpose_dim = self.__class__.get_fullpose_dim()
        if isinstance(fullpose, np.ndarray):
            fullpose_np = fullpose.reshape(-1, fullpose_dim, 3)
        elif isinstance(fullpose, torch.Tensor):
            fullpose_np = fullpose.detach().cpu().numpy().reshape(
                -1, fullpose_dim, 3)
        else:
            self.logger.error('Type of fullpose is not correct.\n' +
                              f'Type: {type(fullpose)}.')
            raise TypeError
        super().__setitem__('fullpose', fullpose_np)

    def set_transl(self, transl: Union[np.ndarray, torch.Tensor]) -> None:
        """Set translation data.

        Args:
            transl (Union[np.ndarray, torch.Tensor]):
                Translation in ndarray or tensor,
                in shape [batch_size, 3].

        Raises:
            TypeError: Type of transl is not correct.
        """
        if isinstance(transl, np.ndarray):
            transl_np = transl.reshape(-1, 3)
        elif isinstance(transl, torch.Tensor):
            transl_np = transl.detach().cpu().numpy().reshape(-1, 3)
        else:
            self.logger.error('Type of transl is not correct.\n' +
                              f'Type: {type(transl)}.')
            raise TypeError
        super().__setitem__('transl', transl_np)

    def set_betas(self, betas: Union[np.ndarray, torch.Tensor]) -> None:
        """Set betas data.

        Args:
            betas (Union[np.ndarray, torch.Tensor]):
                Body shape parameters in ndarray or tensor,
                in shape [batch_size, n].
                n stands for any positive int, typically it's 10.

        Raises:
            TypeError: Type of betas is not correct.
        """
        if isinstance(betas, torch.Tensor):
            betas = betas.detach().cpu().numpy()
        elif not isinstance(betas, np.ndarray):
            self.logger.error('Type of betas is not correct.\n' +
                              f'Type: {type(betas)}.')
            raise TypeError
        if len(betas.shape) == 1:
            betas = betas[np.newaxis, ...]
        betas_dim = betas.shape[-1]
        betas_np = betas.reshape(-1, betas_dim)
        super().__setitem__('betas', betas_np)

    def __setitem__(self, __k: Any, __v: Any) -> None:
        """Set item according to its key.

        Args:
            __k (Any): Key in dict.
            __v (Any): Value in dict.
        """
        if __k == 'gender':
            self.set_gender(__v)
        elif __k == 'transl':
            self.set_transl(__v)
        elif __k == 'fullpose':
            self.set_fullpose(__v)
        elif __k == 'betas':
            self.set_betas(__v)
        else:
            super().__setitem__(__k, __v)

    def get_batch_size(self) -> int:
        """Get batch size.

        Returns:
            int: batch size of fullpose.
        """
        return self.__getitem__('fullpose').shape[0]

    def get_fullpose(self) -> np.ndarray:
        """Get fullpose.

        Returns:
            ndarray: fullpose in shape [batch_size, fullpose_dim, 3].
        """
        fullpose = self.__getitem__('fullpose')
        return fullpose

    def get_global_orient(self) -> np.ndarray:
        """Get global_orient.

        Returns:
            ndarray: global_orient in shape [batch_size, 3].
        """
        fullpose = self.get_fullpose()
        global_orient = fullpose[:, 0].reshape(-1, 3)
        return global_orient

    def get_body_pose(self) -> np.ndarray:
        """Get body_pose.

        Returns:
            ndarray: body_pose in shape [batch_size, BODY_POSE_LEN, 3].
        """
        fullpose = self.get_fullpose()
        batch_size = self.get_batch_size()
        start_idx = 1
        body_pose = fullpose[:, start_idx:start_idx +
                             self.__class__.BODY_POSE_LEN].reshape(
                                 batch_size, -1, 3)
        return body_pose

    def get_transl(self) -> np.ndarray:
        """Get translation.

        Returns:
            ndarray: translation in shape [batch_size, 3].
        """
        return self.__getitem__('transl').reshape(-1, 3)

    def get_betas(self, repeat_betas: bool = True) -> np.ndarray:
        """Get betas.

        Args:
            repeat_betas (bool, optional):
                Whether to repeat betas when its first dim doesn't match
                batch_size. Defaults to True.

        Returns:
            ndarray:
                betas in shape [batch_size, betas_dims] or
                [1, betas_dims].
        """
        batch_size = self.get_batch_size()
        betas = self.__getitem__('betas')
        if repeat_betas and\
                betas.shape[0] == 1 and\
                betas.shape[0] != batch_size:
            betas = betas.repeat(repeats=batch_size, axis=0)
        return betas

    def to_param_dict(self, repeat_betas: bool = True) -> dict:
        """Split fullpose into global_orient and body_pose, return all the
        necessary parameters in one dict.

        Args:
            repeat_betas (bool, optional):
                Whether to repeat betas when its first dim doesn't match
                batch_size. Defaults to True.

        Returns:
            dict:
                A dict of SMPL data, whose keys are
                betas, body_pose, global_orient and transl.
        """
        body_pose = self.get_body_pose().reshape(self.get_batch_size(), -1)
        global_orient = self.get_global_orient().reshape(
            self.get_batch_size(), 3)
        transl = self.get_transl()
        betas = self.get_betas(repeat_betas=repeat_betas)
        dict_to_return = {
            'betas': betas,
            'body_pose': body_pose,
            'global_orient': global_orient,
            'transl': transl,
        }
        return dict_to_return

    def to_tensor_dict(self,
                       repeat_betas: bool = True,
                       device: Union[torch.device, str] = 'cpu') -> dict:
        """It is almost same as self.to_param_dict, but all the values are
        tensors in a specified device. Split fullpose into global_orient and
        body_pose, return all the necessary parameters in one dict.

        Args:
            repeat_betas (bool, optional):
                Whether to repeat betas when its first dim doesn't match
                batch_size. Defaults to True.
            device (Union[torch.device, str], optional):
                A specified device. Defaults to CPU_DEVICE. Defaults to 'cpu'.

        Returns:
            dict: A dict of SMPL data, whose keys are
                betas, body_pose, global_orient and transl.
        """
        np_dict = self.to_param_dict(repeat_betas=repeat_betas)
        dict_to_return = {}
        for key, value in np_dict.items():
            dict_to_return[key] = torch.tensor(
                value, device=device, dtype=torch.float32)
        return dict_to_return

    def dump(self, npz_path: str, overwrite: bool = True):
        """Dump keys and items to an npz file.

        Args:
            npz_path (str):
                Path to a dumped npz file.
            overwrite (bool, optional):
                Whether to overwrite if there is already a file.
                Defaults to True.

        Raises:
            ValueError:
                npz_path does not end with '.npz'.
            FileExistsError:
                When overwrite is False and file exists.
        """
        if not check_path_suffix(npz_path, ['.npz']):
            self.logger.error('Not an npz file.\n' + f'npz_path: {npz_path}')
            raise ValueError
        if not overwrite:
            if check_path_existence(npz_path, 'file') == Existence.FileExist:
                self.logger.error(
                    'File exists while overwrite option not checked.\n' +
                    f'npz_path: {npz_path}')
                raise FileExistsError
        np.savez_compressed(npz_path, **self)

    def from_param_dict(self, smpl_dict: dict) -> None:
        """Load SMPL parameters from smpl_dict.

        Args:
            smpl_dict (dict):
                A dict of ndarray|Tensor parameters.
                global_orient and body_pose are necessary,
                transl and betas are optional.
                Other keys are ignored.


        Raises:
            KeyError: missing necessary keys.
        """
        if not self.__class__.BODY_POSE_KEYS.issubset(smpl_dict):
            self.logger.error('Keys are not enough.\n' +
                              f'smpl_dict\'s keys: {smpl_dict.keys()}')
            raise KeyError
        global_orient = smpl_dict['global_orient']
        batch_size = global_orient.shape[0] \
            if len(global_orient.shape) > 1 else 1
        global_orient = smpl_dict['global_orient'].reshape(batch_size, 3)
        body_pose = smpl_dict['body_pose'].reshape(batch_size, -1)
        if isinstance(global_orient, torch.Tensor):

            def concat_func(data_list, dim):
                return torch.cat(data_list, dim=dim)
        elif isinstance(global_orient, np.ndarray):

            def concat_func(data_list, dim):
                return np.concatenate(data_list, axis=dim)

        fullpose = concat_func([global_orient, body_pose], dim=1)
        self.set_fullpose(fullpose)
        if 'transl' in smpl_dict:
            self.set_transl(smpl_dict['transl'])
        if 'betas' in smpl_dict:
            self.set_betas(smpl_dict['betas'])

    def load(self, npz_path: str):
        """Load data from npz_path and update them to self.

        Args:
            npz_path (str):
                Path to a dumped npz file.
        """
        with np.load(npz_path, allow_pickle=True) as npz_file:
            tmp_data_dict = dict(npz_file)
            for key, value in tmp_data_dict.items():
                self.__setitem__(key, value)



from mmcv.utils import Registry

from .smplify import SMPLify
from .smplifyx import SMPLifyX
from .smplifyxd import SMPLifyXD

REGISTRANTS = Registry('registrant')
REGISTRANTS.register_module(name='SMPLify', module=SMPLify)
REGISTRANTS.register_module(name='SMPLifyX', module=SMPLifyX)
REGISTRANTS.register_module(name='SMPLifyXD', module=SMPLifyXD)


def build_registrant(cfg) -> SMPLify:
    """Build registrant."""
    return REGISTRANTS.build(cfg)



# yapf: disable
import logging
import numpy as np
import torch
from typing import Any, Union
from xrprimer.utils.log_utils import get_logger
from xrprimer.utils.path_utils import (
    Existence, check_path_existence, check_path_suffix,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# yapf: enable


class Keypoints(dict):
    """A class for multi-frame, multi-person keypoints data, based on python
    dict.

    keypoints, mask and convention are the three necessary keys, and we advise
    you to just call Keypoints(). If you'd like to set them manually, it is
    recommended to obey the following turn: convention -> keypoints -> mask.
    """

    def __init__(self,
                 src_dict: dict = None,
                 dtype: Literal['torch', 'numpy', 'auto'] = 'auto',
                 kps: Union[np.ndarray, torch.Tensor, None] = None,
                 mask: Union[np.ndarray, torch.Tensor, None] = None,
                 convention: Union[str, None] = None,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Construct a Keypoints instance with pre-set values. If any of kps,
        mask, convention is provided, it will override the item in src_dict.

        Args:
            src_dict (dict, optional):
                A dict with items in Keypoints fashion.
                Defaults to None.
            dtype (Literal['torch', 'numpy', 'auto'], optional):
                The data type of this Keypoints instance, values will
                be converted to the certain dtype when setting. If
                dtype==auto, it be changed the first time set_keypoints()
                is called, and never changes.
                Defaults to 'auto'.
            kps (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for keypoints,
                kps2d in shape [n_frame, n_person, n_kps, 3],
                kps3d in shape [n_frame, n_person, n_kps, 4].
                Shape [n_kps, 3 or 4] is also accepted, unsqueezed
                automatically. Defaults to None.
            mask (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for keypoint mask,
                in shape [n_frame, n_person, n_kps],
                in dtype uint8.
                Shape [n_kps, ] is also accepted, unsqueezed
                automatically. Defaults to None.
            convention (str, optional):
                Convention name of the keypoints,
                can be found in KEYPOINTS_FACTORY.
                Defaults to None.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        if src_dict is not None:
            super().__init__(src_dict)
        else:
            super().__init__()
        self.logger = get_logger(logger)

        if dtype == 'auto':
            if kps is not None:
                dtype = __get_array_type_str__(kps, logger)
            elif src_dict is not None and 'keypoints' in src_dict:
                dtype = __get_array_type_str__(src_dict['keypoints'], logger)
        self.dtype = dtype

        if convention is not None:
            self.set_convention(convention)

        if kps is not None:
            self.set_keypoints(kps)

        if mask is None and 'mask' not in self and\
                'keypoints' in self:
            default_n_kps = self.get_keypoints_number()
            mask = np.ones(shape=(default_n_kps, ))
        if mask is not None:
            self.set_mask(mask)

    @classmethod
    def fromfile(cls, npz_path: str) -> 'Keypoints':
        """Construct a body model data structure from an npz file.

        Args:
            npz_path (str):
                Path to a dumped npz file.

        Returns:
            Keypoints:
                A Keypoints instance load from file.
        """
        ret_instance = cls()
        ret_instance.load(npz_path)
        return ret_instance

    def set_keypoints(self, kps: Union[np.ndarray, torch.Tensor]) -> None:
        """Set keypoints array.

        Args:
            kps (Union[np.ndarray, torch.Tensor]):
                A tensor or ndarray for keypoints,
                kps2d in shape [n_frame, n_person, n_kps, 3],
                kps3d in shape [n_frame, n_person, n_kps, 4].
                Shape [n_kps, 3 or 4] is also accepted, unsqueezed
                automatically.

        Raises:
            TypeError: Type of keypoints is wrong.
            ValueError: kps.shape[-1] is wrong.
            ValueError: Shape of kps is wrong.
        """
        if self.dtype == 'auto':
            self.dtype = __get_array_type_str__(kps, self.logger)
        keypoints = __get_array_in_type__(
            array=kps, type=self.dtype, logger=self.logger)
        # shape: frame_n, person_n, kp_n, dim+score
        if keypoints.shape[-1] not in (3, 4):
            self.logger.error('shape[-1] of kps2d should be 3,' +
                              ' shape[-1] of kps3d should be 4.\n' +
                              f'kps.shape[-1]: {kps.shape[-1]}.')
            raise ValueError
        if len(keypoints.shape) == 2:
            keypoints = keypoints.reshape(1, 1, keypoints.shape[0],
                                          keypoints.shape[1])
        if len(keypoints.shape) != 4:
            self.logger.error('Shape of keypoints should be' +
                              ' [n_frame, n_person, n_kps, dim+1].\n' +
                              f'kps.shape: {kps.shape}.')
            raise ValueError
        super().__setitem__('keypoints', keypoints)

    def set_convention(self, convention: str) -> None:
        """Set convention name of the keypoints.

        Args:
            convention (str):
                Convention name of the keypoints,
                can be found in KEYPOINTS_FACTORY.

        Raises:
            TypeError: Type of convention is not str.
        """
        if not isinstance(convention, str):
            self.logger.error('Type of convention is not str.\n' +
                              f'type(convention): {type(convention)}.')
            raise TypeError
        super().__setitem__('convention', convention)

    def set_mask(self, mask: Union[np.ndarray, torch.Tensor]) -> None:
        """Set mask of the keypoints. It should be called after the
        corresponding keypoints has been set.

        Args:
            mask (Union[np.ndarray, torch.Tensor]):
                A tensor or ndarray for keypoint mask,
                in shape [n_frame, n_person, n_kps],
                in dtype uint8.
                Shape [n_kps, ] is also accepted, unsqueezed
                automatically.

        Raises:
            TypeError: Type of mask is wrong.
            ValueError: Shape of mask is wrong.
        """
        if self.dtype == 'auto':
            self.dtype = __get_array_type_str__(mask, self.logger)

        mask = __get_array_in_type__(
            array=mask, type=self.dtype, logger=self.logger)

        if self.dtype == 'torch':

            def to_type_uint8_func(data):
                return data.to(dtype=torch.uint8)

        else:

            def to_type_uint8_func(data):
                return data.astype(np.uint8)

        mask = to_type_uint8_func(mask)
        keypoints_shape = self.get_keypoints().shape
        if len(mask.shape) == 1:
            mask = mask.reshape(1, 1, len(mask))
            mask = mask.repeat(keypoints_shape[0], axis=0)
            mask = mask.repeat(keypoints_shape[1], axis=1)
        if len(mask.shape) != 3 or \
                mask.shape != keypoints_shape[:3]:
            self.logger.error('Shape of mask should be' +
                              ' [n_frame, n_person, n_kps].\n' +
                              f'mask.shape: {mask.shape}.' +
                              f'keypoints.shape: {keypoints_shape}.')
            raise ValueError
        super().__setitem__('mask', mask)

    def __setitem__(self, __k: Any, __v: Any) -> None:
        """Set item according to its key.

        Args:
            __k (Any): Key in dict.
            __v (Any): Value in dict.
        """
        if __k == 'keypoints':
            self.set_keypoints(__v)
        elif __k == 'convention':
            self.set_convention(__v)
        elif __k == 'mask':
            self.set_mask(__v)
        else:
            super().__setitem__(__k, __v)

    def get_keypoints(self) -> Union[np.ndarray, torch.Tensor]:
        """Get keypoints array.

        Returns:
            np.ndarray: keypoints
        """
        return self['keypoints']

    def get_mask(self) -> Union[np.ndarray, torch.Tensor]:
        """Get keypoints mask.

        Returns:
            np.ndarray: mask
        """
        return self['mask']

    def get_convention(self) -> str:
        """Get keypoints convention name.

        Returns:
            str: convention
        """
        return self['convention']

    def get_frame_number(self) -> int:
        """Get frame number of keypoints.

        Returns:
            int: frame number
        """
        return self.get_keypoints().shape[0]

    def get_person_number(self) -> int:
        """Get person number of keypoints.

        Returns:
            int: person number
        """
        return self.get_keypoints().shape[1]

    def get_keypoints_number(self) -> int:
        """Get number of keypoints.

        Returns:
            int: keypoints number
        """
        return self.get_keypoints().shape[2]

    def to_tensor(self,
                  device: Union[torch.device, str] = 'cpu') -> 'Keypoints':
        """Return all the necessary values for keypoints expression in another
        Keypoints instance, convert ndarray into Tensor.

        Args:
            device (Union[torch.device, str], optional):
                A specified device.
                Defaults to 'cpu'.

        Returns:
            Keypoints: An instance of Keypoints data, whose keys are
                keypoints, mask, convention.
        """
        kps_to_return = self.__class__(
            dtype='torch',
            kps=self.get_keypoints(),
            mask=self.get_mask(),
            convention=self.get_convention(),
            logger=self.logger)
        kps_to_return.set_keypoints(kps_to_return.get_keypoints().to(device))
        kps_to_return.set_mask(kps_to_return.get_mask().to(device))
        return kps_to_return

    def to_numpy(self, ) -> 'Keypoints':
        """Return all the necessary values for keypoints expression in another
        Keypoints instance, convert Tensor into numpy.

        Returns:
            Keypoints: An instance of Keypoints data, whose keys are
                keypoints, mask, convention.
        """
        kps_to_return = self.__class__(
            dtype='numpy',
            kps=self.get_keypoints(),
            mask=self.get_mask(),
            convention=self.get_convention(),
            logger=self.logger)
        return kps_to_return

    def dump(self, npz_path: str, overwrite: bool = True):
        """Dump keys and items to an npz file.

        Args:
            npz_path (str):
                Path to a dumped npz file.
            overwrite (bool, optional):
                Whether to overwrite if there is already a file.
                Defaults to True.

        Raises:
            ValueError:
                npz_path does not end with '.npz'.
            FileExistsError:
                When overwrite is False and file exists.
        """
        if not check_path_suffix(npz_path, ['.npz']):
            self.logger.error('Not an npz file.\n' + f'npz_path: {npz_path}')
            raise ValueError
        if not overwrite:
            if check_path_existence(npz_path, 'file') == Existence.FileExist:
                self.logger.error(
                    'File exists while overwrite option not checked.\n' +
                    f'npz_path: {npz_path}')
                raise FileExistsError
        if self.dtype == 'numpy':
            dict_to_save = self
        else:  # else self.dtype == tensor
            dict_to_save = self.to_numpy()
        np.savez_compressed(npz_path, **dict_to_save)

    def load(self, npz_path: str):
        """Load data from npz_path and update them to self.

        Args:
            npz_path (str):
                Path to a dumped npz file.
        """
        with np.load(npz_path, allow_pickle=True) as npz_file:
            tmp_data_dict = dict(npz_file)
            for key, value in tmp_data_dict.items():
                if isinstance(value, np.ndarray) and\
                        len(value.shape) == 0:
                    # value is not an ndarray before dump
                    value = value.item()
                self.__setitem__(key, value)

    def clone(self) -> 'Keypoints':
        """Clone a Keypoints instance as self.

        Returns:
            Keypoints:
                A deep copied instance of Keypoints,
                with the same dtype and value as self.
        """
        ret_kps = self.__class__(
            dtype=self.dtype,
            kps=__copy_array_tensor__(self.get_keypoints()),
            mask=__copy_array_tensor__(self.get_mask()),
            convention=self.get_convention(),
            logger=self.logger)
        return ret_kps


def __get_array_type_str__(array, logger) -> Literal['torch', 'numpy']:
    if isinstance(array, torch.Tensor):
        return 'torch'
    elif isinstance(array, np.ndarray):
        return 'numpy'
    else:
        logger = get_logger(logger)
        logger.error('Type of array is not correct.\n' +
                     f'Type: {type(array)}.')
        raise TypeError


def __get_array_in_type__(array: Union[torch.Tensor, np.ndarray],
                          type: Literal['torch', 'numpy'],
                          logger: Union[None, str, logging.Logger]):
    logger = get_logger(logger)
    if type == 'numpy':
        if isinstance(array, torch.Tensor):
            array = array.detach().cpu().numpy()
        elif not isinstance(array, np.ndarray):
            logger.error('Type of array is not correct.\n' +
                         f'Type: {type(array)}.')
    else:  # type == 'torch'
        if isinstance(array, np.ndarray):
            array = torch.from_numpy(array)
        elif not isinstance(array, torch.Tensor):
            logger.error('Type of array is not correct.\n' +
                         f'Type: {type(array)}.')
    return array


def __copy_array_tensor__(
        data: Union[np.ndarray,
                    torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    if isinstance(data, np.ndarray):
        return data.copy()
    else:
        return data.clone()



import logging
import numpy as np
import torch
from typing import Union
from xrprimer.utils.log_utils import get_logger

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def convert_bbox(
    data: Union[torch.Tensor, np.ndarray],
    src: Literal['xyxy', 'xywh'],
    dst: Literal['xyxy', 'xywh'],
    logger: Union[None, str, logging.Logger] = None
) -> Union[torch.Tensor, np.ndarray]:
    """Convert bbox2d following the mapping correspondence between src and dst
    keypoints definition. Supported conventions by now: xyxy, xywh.

    Args:
        data (Union[torch.Tensor, np.ndarray]):
            Bbox2d data. The last dim should be (x, y, x, y, score),
            or (x, y, w, h, score), and score is optional.
        src (str):
            Bbox type, xyxy or xywh.
        dst (str):
            Bbox type, xyxy or xywh.
        logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

    Raises:
        ValueError: Either src or dst is not in defined conventions.

    Returns:
        Union[torch.Tensor, np.ndarray]:
            The converted bbox2d in the same shape and type as input data.
    """
    logger = get_logger(logger)
    if src == dst:
        logger.warning('Src convention equals to dst convention.')
        return data
    if isinstance(data, torch.Tensor):

        def abs_func(data):
            return torch.abs(input=data)

        def min_func(data):
            return torch.min(input=data, dim=-1, keepdim=False)[0]

        def clone_func(data):
            return torch.clone(data)
    elif isinstance(data, np.ndarray):

        def abs_func(data):
            return np.absolute(data)

        def min_func(data):
            return np.min(data, axis=-1, keepdims=False)

        def clone_func(data):
            return np.copy(data)

    if src == 'xyxy':
        w = abs_func(data[..., 0] - data[..., 2])
        h = abs_func(data[..., 1] - data[..., 3])
        x_min = min_func(data[..., slice(0, 4, 2)])
        y_min = min_func(data[..., slice(1, 4, 2)])
    elif src == 'xywh':
        w = data[..., 2]
        h = data[..., 3]
        x_min = data[..., 0]
        y_min = data[..., 1]
    else:
        raise ValueError('Wrong source name.')
    output = clone_func(data)
    output[..., 0] = x_min
    output[..., 1] = y_min
    if dst == 'xyxy':
        output[..., 2] = x_min + w
        output[..., 3] = y_min + h
    elif dst == 'xywh':
        output[..., 2] = w
        output[..., 3] = h
    else:
        raise ValueError('Wrong destination name.')
    return output



import numpy as np
import torch
from abc import ABCMeta
from mmcv.cnn.resnet import ResNet
from torch.autograd import Variable
from torch.nn import functional as F
from typing import Optional, Union

from .base_architecture import BaseArchitecture


def pairwise_affinity(query_features):
    m = query_features.size(0)

    dist = torch.pow(query_features, 2).sum(
        dim=1, keepdim=True).expand(m, m) + torch.pow(query_features, 2).sum(
            dim=1, keepdim=True).expand(m, m).t()
    dist.addmm_(query_features, query_features.t(), beta=1, alpha=-2)
    normalized_affinity = -(dist - dist.mean()) / dist.std()
    affinity = torch.sigmoid(normalized_affinity * torch.tensor(5.))
    return affinity


def reranking(query_features, k1=20, k2=6, lamda_value=0.3):
    feat = torch.cat((query_features, query_features))
    n_query, n_all = query_features.size(0), feat.size(0)
    feat = feat.view(n_all, -1)

    dist = torch.pow(feat, 2).sum(dim=1, keepdim=True).expand(n_all, n_all)
    dist = dist + dist.t()
    dist.addmm_(feat, feat.t(), beta=1, alpha=-2)

    original_dist = dist.cpu().numpy()
    n_all = original_dist.shape[0]
    original_dist = np.transpose(original_dist /
                                 np.max(original_dist, axis=0) + 1e-9)
    V = np.zeros_like(original_dist).astype(np.float16)
    initial_rank = np.argsort(original_dist).astype(np.int32)

    for i in range(n_all):
        # k-reciprocal neighbors
        forward_k_neigh_index = initial_rank[i, :k1 + 1]
        backward_k_neigh_index = initial_rank[forward_k_neigh_index, :k1 + 1]
        fi = np.where(backward_k_neigh_index == i)[0]
        k_reciprocal_index = forward_k_neigh_index[fi]
        k_reciprocal_expansion_index = k_reciprocal_index
        for j in range(len(k_reciprocal_index)):
            candidate = k_reciprocal_index[j]
            candidate_forward_k_neigh_index = initial_rank[
                candidate, :int(np.around(k1 / 2)) + 1]
            candidate_backward_k_neigh_index = initial_rank[
                candidate_forward_k_neigh_index, :int(np.around(k1 / 2)) + 1]
            fi_candidate = np.where(
                candidate_backward_k_neigh_index == candidate)[0]
            candidate_k_reciprocal_index = \
                candidate_forward_k_neigh_index[fi_candidate]
            if len(
                    np.intersect1d(candidate_k_reciprocal_index,
                                   k_reciprocal_index)
            ) > 2 / 3 * len(candidate_k_reciprocal_index):
                k_reciprocal_expansion_index = np.append(
                    k_reciprocal_expansion_index, candidate_k_reciprocal_index)

        k_reciprocal_expansion_index = np.unique(k_reciprocal_expansion_index)
        weight = np.exp(-original_dist[i, k_reciprocal_expansion_index])
        V[i, k_reciprocal_expansion_index] = weight / np.sum(weight)
    original_dist = original_dist[:n_query, ]
    if k2 != 1:
        V_query_expansion = np.zeros_like(V, dtype=np.float16)
        for i in range(n_all):
            V_query_expansion[i, :] = \
                np.mean(V[initial_rank[i, :k2], :], axis=0)
        V = V_query_expansion
    invIndex = []
    for i in range(n_all):
        invIndex.append(np.where(V[:, i] != 0)[0])

    jaccard_dist = np.zeros_like(original_dist, dtype=np.float16)

    for i in range(n_query):
        temp_min = np.zeros(shape=[1, n_all], dtype=np.float16)
        indNonZero = np.where(V[i, :] != 0)[0]
        indImages = []
        indImages = [invIndex[ind] for ind in indNonZero]
        for j in range(len(indNonZero)):
            temp_min[0, indImages[j]] = temp_min[0, indImages[j]] + \
                np.minimum(V[i, indNonZero[j]], V[indImages[j], indNonZero[j]])
        jaccard_dist[i] = 1 - temp_min / (2 - temp_min)

    final_dist = jaccard_dist * (1 - lamda_value) + original_dist * lamda_value
    final_dist = final_dist[:n_query, n_query:]
    final_dist = torch.from_numpy(final_dist)
    return final_dist


class AppearanceAffinityEstimator(BaseArchitecture, metaclass=ABCMeta):
    """AppearanceAffinityEstimator Architecture."""

    def __init__(self, init_cfg: Optional[Union[list, dict, None]] = None):
        super(AppearanceAffinityEstimator, self).__init__(init_cfg)
        self.backbone = ResNet(depth=50, out_indices=[3])

    def extract_features(self, image_tensor):
        image_tensor = Variable(image_tensor)
        outputs = self.forward_test(image_tensor)
        features = outputs.data
        return features

    def get_affinity(self,
                     image_tensor: torch.Tensor,
                     rerank=False) -> torch.Tensor:
        """Compute affinity matrix.

        Args:
            image_tensor (torch.Tensor)
            rerank (bool, optional): Defaults to False.

        Returns:
            torch.Tensor: affinity
        """
        query_features = self.extract_features(image_tensor)
        if rerank:
            affinity = reranking(query_features)
        else:
            affinity = pairwise_affinity(query_features)
        return affinity

    def forward_train(self, **kwargs):
        """Forward function for general training.

        For 3d pose estimation, we do not use this interface.
        """
        raise NotImplementedError('This interface should not be used in '
                                  'current training schedule. Please use '
                                  '`train_step` for training.')

    def forward_test(self, x: torch.Tensor, **kwargs):
        assert len(kwargs) == 0
        for _, module in self.backbone._modules.items():
            x = module(x)

        # if self.cut_at_pooling:
        #     return x

        x = F.avg_pool2d(x, x.size()[2:])
        x = x.view(x.size(0), -1)

        x = F.normalize(x)
        return x



import cv2
import logging
import numpy as np
import torch
from typing import Union
from xrprimer.utils.log_utils import get_logger

from xrmocap.transform.convention.bbox_convention import convert_bbox


def compute_similarity_transform(X: np.ndarray,
                                 Y: np.ndarray,
                                 compute_optimal_scale=False):
    """A port of MATLAB's `procrustes` function to Numpy. Adapted from
    http://stackoverflow.com/a/18927641/1884420.

    Args:
        X (np.ndarray): Array NxM of targets, with N number of points and
           M point dimensionality.
        Y (np.ndarray): Array NxM of inputs
        compute_optimal_scale (bool, optional): whether we compute optimal
            scale or force it to be 1. Defaults to False.

    Returns:
        d: squared error after transformation
        Z (np.ndarray): transformed Y
        T (np.ndarray): computed rotation
        b: scaling
        c: translation
    """
    muX = X.mean(0)
    muY = Y.mean(0)

    X0 = X - muX
    Y0 = Y - muY

    ssX = (X0**2.).sum()
    ssY = (Y0**2.).sum()

    # centred Frobenius norm
    normX = np.sqrt(ssX)
    normY = np.sqrt(ssY)

    # scale to equal (unit) norm
    X0 = X0 / normX
    Y0 = Y0 / normY

    # optimum rotation matrix of Y
    A = np.dot(X0.T, Y0)
    U, s, Vt = np.linalg.svd(A, full_matrices=False)
    V = Vt.T
    T = np.dot(V, U.T)

    # Make sure we have a rotation
    detT = np.linalg.det(T)
    V[:, -1] *= np.sign(detT)
    s[-1] *= np.sign(detT)
    T = np.dot(V, U.T)

    traceTA = s.sum()

    if compute_optimal_scale:  # Compute optimum scaling of Y.
        b = traceTA * normX / normY
        d = 1 - traceTA**2
        Z = normX * traceTA * np.dot(Y0, T) + muX
    else:  # If no scaling allowed
        b = 1
        d = 1 + ssY / ssX - 2 * traceTA * normY / normX
        Z = normY * np.dot(Y0, T) + muX

    c = muX - b * np.dot(muY, T)
    return d, Z, T, b, c


def get_affine_transform(center,
                         scale,
                         rot,
                         output_size,
                         shift=np.array([0, 0], dtype=np.float32),
                         inv=0):
    """This function calculates affine transformation from given center, scale,
    rotation and shift."""

    def get_3rd_point(a, b):
        # get the third point from the first two points to define
        # an affine transformation
        direct = a - b
        return np.array(b) + np.array([-direct[1], direct[0]],
                                      dtype=np.float32)

    if isinstance(scale, torch.Tensor):
        scale = np.array(scale.cpu())
    if isinstance(center, torch.Tensor):
        center = np.array(center.cpu())
    if not isinstance(scale, np.ndarray) and not isinstance(scale, list):
        scale = np.array([scale, scale])

    scale_tmp = scale * 200.0
    src_w, src_h = scale_tmp[0], scale_tmp[1]
    dst_w, dst_h = output_size[0], output_size[1]

    rot_rad = np.pi * rot / 180
    if src_w >= src_h:
        src_dir = get_direction([0, src_w * -0.5], rot_rad)
        dst_dir = np.array([0, dst_w * -0.5], np.float32)
    else:
        src_dir = get_direction([src_h * -0.5, 0], rot_rad)
        dst_dir = np.array([dst_h * -0.5, 0], np.float32)

    src = np.zeros((3, 2), dtype=np.float32)
    dst = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale_tmp * shift  # x,y
    src[1, :] = center + src_dir + scale_tmp * shift
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir

    src[2:, :] = get_3rd_point(src[0, :], src[1, :])
    dst[2:, :] = get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    return trans


def get_direction(src_point, rot_rad):
    sn, cs = np.sin(rot_rad), np.cos(rot_rad)

    src_result = [0, 0]
    src_result[0] = src_point[0] * cs - src_point[1] * sn
    src_result[1] = src_point[0] * sn + src_point[1] * cs

    return src_result


def get_scale(image_size, resized_size):
    w, h = image_size
    w_resized, h_resized = resized_size
    if w / w_resized < h / h_resized:
        w_pad = h / h_resized * w_resized
        h_pad = h
    else:
        w_pad = w
        h_pad = w / w_resized * h_resized
    scale = np.array([w_pad / 200.0, h_pad / 200.0], dtype=np.float32)

    return scale


def project_3dpts(X, K, R, t, Kd):
    """Projects points X (3xN) using camera intrinsics K (3x3), extrinsics
    (R,t) and distortion parameters Kd=[k1,k2,p1,p2,k3].

    Roughly, x = K*(R*X + t) + distortion
    See http://docs.opencv.org/2.4/doc/tutorials/
    calib3d/camera_calibration/camera_calibration.html
    or cv2.projectPoints
    """

    x = torch.mm(R, X) + t
    # x = np.dot(R, X) + t

    x[0:2, :] = x[0:2, :] / (x[2, :] + 1e-5)

    r = x[0, :] * x[0, :] + x[1, :] * x[1, :]

    x[0, :] = (
        x[0, :] * (1 + Kd[0] * r + Kd[1] * r * r + Kd[4] * r * r * r) +
        2 * Kd[2] * x[0, :] * x[1, :] + Kd[3] * (r + 2 * x[0, :] * x[0, :]))
    x[1, :] = (
        x[1, :] * (1 + Kd[0] * r + Kd[1] * r * r + Kd[4] * r * r * r) +
        2 * Kd[3] * x[0, :] * x[1, :] + Kd[2] * (r + 2 * x[1, :] * x[1, :]))

    x[0, :] = K[0, 0] * x[0, :] + K[0, 1] * x[1, :] + K[0, 2]
    x[1, :] = K[1, 0] * x[0, :] + K[1, 1] * x[1, :] + K[1, 2]

    return x


def compute_iou(
    rec_1: Union[np.ndarray, torch.Tensor],
    rec_2: Union[np.ndarray, torch.Tensor],
    bbox_convention: str = 'xyxy',
    logger: Union[None, str, logging.Logger] = None
) -> Union[np.float64, torch.Tensor]:
    """Compute iou.

    Args:
        rec_1 (Union[np.ndarray, torch.Tensor]):
            Bbox2d data. The last dim should be
            (x, y, x, y, score), or (x, y, w, h, score), and score is optional.
        rec_2 (Union[np.ndarray, torch.Tensor]):
            Bbox2d data. The last dim should be
            (x, y, x, y, score), or (x, y, w, h, score), and score is optional.
        bbox_convention (str, optional): Bbox type, xyxy or xywh.
            Defaults to 'xyxy'.
        logger (Union[None, str, logging.Logger], optional):
            Logger for logging. If None, root logger will be selected.
            Defaults to None.
            (rec_1)
        1--------1
        1   1----1------1
        1---1----1      1
            1           1
            1-----------1 (rec_2)
    Returns:
        Union[np.float64, torch.Tensor]: iou
    """
    logger = get_logger(logger)
    assert bbox_convention == 'xywh' or bbox_convention == 'xyxy'
    if bbox_convention == 'xywh':
        rec_1 = convert_bbox(rec_1, src='xywh', dst='xyxy')
        rec_2 = convert_bbox(rec_2, src='xywh', dst='xyxy')

    s_rec1 = (rec_1[2] - rec_1[0]) * (rec_1[3] - rec_1[1])
    s_rec2 = (rec_2[2] - rec_2[0]) * (rec_2[3] - rec_2[1])
    sum_s = s_rec1 + s_rec2
    left = max(rec_1[0], rec_2[0])
    right = min(rec_1[2], rec_2[2])
    bottom = max(rec_1[1], rec_2[1])
    top = min(rec_1[3], rec_2[3])
    if left >= right or top <= bottom:
        if isinstance(rec_1, torch.Tensor):
            return torch.tensor(0)
        else:
            return np.float64(0)
    else:
        inter = (right - left) * (top - bottom)
        iou = (inter / (sum_s - inter)) * 1.0
        return iou
